# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import testtools

from pypowervm import adapter as adpt
import pypowervm.entities as ent
from pypowervm import exceptions as pvm_exc
from pypowervm.tasks import network_bridger as net_br
import pypowervm.tests.test_fixtures as fx
from pypowervm.tests.test_utils import pvmhttp
from pypowervm.wrappers import managed_system as pvm_ms
from pypowervm.wrappers import network as pvm_net

MGR_NET_BR_FAILOVER_FILE = 'nbbr_network_bridge_failover.txt'
MGR_NET_BR_FILE = 'nbbr_network_bridge.txt'
MGR_NET_BR_PEER_FILE = 'nbbr_network_bridge_peer.txt'
MGR_VNET_FILE = 'nbbr_virtual_network.txt'
MGR_VSW_FILE = 'nbbr_virtual_switch.txt'
ORPHAN_VIOS_FEED = 'fake_vios_feed.txt'
ORPHAN_CNA_FEED = 'cna_feed.txt'


class TestNetworkBridger(testtools.TestCase):
    """General tests for the Network Bridger superclass.

    Subclasses of Network Bridgers should extend this class.
    """

    def setUp(self):
        super(TestNetworkBridger, self).setUp()

        self.adptfx = self.useFixture(fx.AdapterFx(
            traits=fx.LocalPVMTraits))
        self.adpt = self.adptfx.adpt

        def resp(file_name):
            return pvmhttp.load_pvm_resp(
                file_name, adapter=self.adpt).get_response()

        self.mgr_nbr_resp = resp(MGR_NET_BR_FILE)
        self.mgr_nbr_fo_resp = resp(MGR_NET_BR_FAILOVER_FILE)
        self.mgr_nbr_peer_resp = resp(MGR_NET_BR_PEER_FILE)
        self.mgr_vnet_resp = resp(MGR_VNET_FILE)
        self.mgr_vsw_resp = resp(MGR_VSW_FILE)
        self.orphan_vio_resp = resp(ORPHAN_VIOS_FEED)
        self.orphan_cna_feed = resp(ORPHAN_CNA_FEED)

        self.host_uuid = 'c5d782c7-44e4-3086-ad15-b16fb039d63b'
        self.nb_uuid = 'b6a027a8-5c0b-3ac0-8547-b516f5ba6151'
        self.nb_uuid_peer = '9af89d52-5892-11e5-885d-feff819cdc9f'

    def test_ensure_vlan_on_nb(self):
        """This does a happy path test.  Assumes VLAN on NB already.

        No subclass invocation.
        """
        self.adpt.read.return_value = self.mgr_nbr_resp
        net_br.ensure_vlan_on_nb(self.adpt, self.host_uuid, self.nb_uuid, 2227)
        self.assertEqual(1, self.adpt.read.call_count)

    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridger'
                '._validate_orphan_on_ensure')
    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridger'
                '._remove_vlan_from_nb_synch')
    def test_ensure_vlan_on_nb_wrong_peer(self, mock_remove, mock_orphan):
        """Test moving vlan from one peer to another.

        No subclass invocation.
        """
        self.adpt.read.side_effect = [
            self.mgr_nbr_peer_resp, self.mgr_vsw_resp, self.mgr_vnet_resp]

        def validate_of_update_nb(*kargs, **kwargs):
            # Validate args
            nb = kargs[0]
            self.assertEqual(self.nb_uuid, nb.uuid)
            return nb.entry

        self.adpt.update_by_path.side_effect = validate_of_update_nb

        net_br.ensure_vlan_on_nb(self.adpt, self.host_uuid, self.nb_uuid, 1001)
        mock_remove.assert_called_once_with(
            self.nb_uuid_peer, 1001, fail_if_pvid=True, existing_nbs=mock.ANY)

    def test_is_arbitrary_vid(self):
        nbs = pvm_net.NetBridge.wrap(self.mgr_nbr_resp)
        bridger = net_br.NetworkBridger(self.adpt, self.host_uuid)
        self.assertTrue(bridger._is_arbitrary_vid(4094, nbs))
        self.assertFalse(bridger._is_arbitrary_vid(2227, nbs))

    def test_find_new_arbitrary_vid(self):
        nbs = pvm_net.NetBridge.wrap(self.mgr_nbr_resp)
        bridger = net_br.NetworkBridger(self.adpt, self.host_uuid)
        self.assertEqual(4093, bridger._find_new_arbitrary_vid(nbs))
        self.assertEqual(4092, bridger._find_new_arbitrary_vid(nbs,
                                                               others=[4093]))

    def test_remove_vlan_from_nb_bad_vid(self):
        """Attempt to remove a VID that can't be taken off NB."""
        # Mock Data
        self.adpt.read.return_value = self.mgr_nbr_resp

        # Should fail if fail_if_pvid set to True
        self.assertRaises(pvm_exc.PvidOfNetworkBridgeError,
                          net_br.remove_vlan_from_nb, self.adpt,
                          self.host_uuid, self.nb_uuid, 2227, True)

        # Should not fail if fail_if_pvid set to False, but shouldn't call
        # update either.
        net_br.remove_vlan_from_nb(self.adpt, self.host_uuid, self.nb_uuid,
                                   '2227')
        self.assertEqual(0, self.adpt.update.call_count)

    def _setup_reassign_arbitrary_vid(self):
        vsw_p = mock.patch('pypowervm.wrappers.network.VSwitch.search')
        self.mock_vsw = vsw_p.start()
        self.addCleanup(vsw_p.stop)
        vnet = pvm_net.VNet._bld(self.adpt).entry
        resp1 = adpt.Response('reqmethod', 'reqpath', 'status', 'reason', {})
        resp1.feed = ent.Feed({}, [vnet])
        self.adpt.read.return_value = resp1
        self.adpt.read_by_href.return_value = vnet
        nb = pvm_net.NetBridge.wrap(self.mgr_nbr_resp)[0]
        resp2 = adpt.Response('reqmethod', 'reqpath', 'status', 'reason', {})
        resp2.entry = nb.entry
        self.adpt.update.return_value = resp2

        vsw = pvm_net.VSwitch.wrap(self.mgr_vsw_resp)[0]
        self.mock_vsw.return_value = vsw
        return nb

    def test_build_orphan_map(self):
        self.adpt.read.side_effect = [self.orphan_vio_resp]
        bridger = net_br.NetworkBridger(self.adpt, self.host_uuid)
        orphan_map = bridger._build_orphan_map()

        expected_map = {
            0: {'nimbus-ch03-p2-vios2': {'ent11': [4092, 2018, 2019]}},
            1: {'nimbus-ch03-p2-vios2': {'ent12': [2800, 2801]}}
        }

        self.assertEqual(expected_map, orphan_map)

    def test_validate_orphan_on_ensure(self):
        """Tests the _validate_orphan_on_ensure method."""
        self.adpt.read.side_effect = [self.orphan_vio_resp]
        bridger = net_br.NetworkBridger(self.adpt, self.host_uuid)

        # Test the Trunk Path - PVID and then an additional
        self.assertRaises(
            pvm_exc.OrphanVLANFoundOnProvision,
            bridger._validate_orphan_on_ensure, 4092, 0)
        self.assertRaises(
            pvm_exc.OrphanVLANFoundOnProvision,
            bridger._validate_orphan_on_ensure, 2018, 0)

        # Different vSwitch
        self.assertRaises(
            pvm_exc.OrphanVLANFoundOnProvision,
            bridger._validate_orphan_on_ensure, 2800, 1)

        # Shouldn't fail on a good vlan
        bridger._validate_orphan_on_ensure(2, 0)
        bridger._validate_orphan_on_ensure(1, 0)
        bridger._validate_orphan_on_ensure(4094, 1)

    def test_get_orphan_vlans(self):
        """Tests the _get_orphan_vlans method."""
        self.adpt.read.side_effect = [self.orphan_vio_resp]
        bridger = net_br.NetworkBridger(self.adpt, self.host_uuid)

        self.assertListEqual([], bridger._get_orphan_vlans(2))
        self.assertListEqual([2018, 2019, 4092],
                             bridger._get_orphan_vlans(0))
        self.assertListEqual([2800, 2801], bridger._get_orphan_vlans(1))


class TestNetworkBridgerVNet(TestNetworkBridger):
    """General tests for the network bridge super class and the VNet impl."""

    def setUp(self):
        super(TestNetworkBridgerVNet, self).setUp()
        self.adptfx.set_traits(fx.RemoteHMCTraits)

    @mock.patch('oslo_concurrency.lockutils.lock')
    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridger.'
                '_get_orphan_vlans')
    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridger.'
                '_validate_orphan_on_ensure')
    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridgerVNET.'
                '_reassign_arbitrary_vid')
    @mock.patch('pypowervm.wrappers.network.NetBridge.supports_vlan')
    def test_ensure_vlan_on_nb_reassign(
            self, mock_support_vlan, mock_reassign, mock_orphan_validate,
            mock_orphans, mock_lock):
        """Validates that after update, we support the VLAN."""
        # Have the response
        self.adpt.read.return_value = self.mgr_nbr_resp

        # First call, say that we don't support the VLAN (which is true).
        # Second call, fake out that we now do.
        # Works in pairs, as there are two VLANs we're working through.
        mock_support_vlan.side_effect = [False, False, True, True]
        mock_orphans.return_value = []

        # Invoke
        net_br.ensure_vlans_on_nb(self.adpt, self.host_uuid, self.nb_uuid,
                                  [4093, 4094])
        self.assertEqual(2, self.adpt.read.call_count)
        self.assertEqual(1, mock_reassign.call_count)
        self.assertEqual(1, mock_lock.call_count)

        # Should be called re-assigning 4094 (old) to 4092.  Shouldn't be
        # 4093 as that is also an additional VLAN.
        mock_reassign.assert_called_once_with(4094, 4092, mock.ANY)

    @mock.patch('oslo_concurrency.lockutils.lock')
    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridger.'
                '_validate_orphan_on_ensure')
    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridgerVNET.'
                '_find_or_create_vnet')
    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridgerVNET.'
                '_is_arbitrary_vid')
    def test_ensure_vlan_on_nb_new_vlan(self, mock_arb_vid, mock_find_vnet,
                                        mock_orphan_validate, mock_lock):
        """Validates new VLAN on existing Load Group."""
        # Build the responses
        self.adpt.read.side_effect = [self.mgr_nbr_resp, self.mgr_vsw_resp,
                                      self.mgr_vnet_resp]
        mock_arb_vid.return_value = False
        mock_vnet = mock.MagicMock()
        mock_vnet.related_href = 'fake_href'
        mock_find_vnet.return_value = mock_vnet

        def validate_of_update_nb(*kargs, **kwargs):
            # Validate args
            nb = kargs[0]
            self.assertIsNotNone(nb)
            self.assertEqual(1,
                             len(nb.load_grps[0].vnet_uri_list))
            self.assertEqual(2,
                             len(nb.load_grps[1].vnet_uri_list))
            self.assertEqual(self.nb_uuid, nb.uuid)

            return nb.entry

        self.adpt.update_by_path.side_effect = validate_of_update_nb

        # Invoke
        net_br.ensure_vlan_on_nb(self.adpt, self.host_uuid, self.nb_uuid, 2000)

        # Validate the calls
        self.assertEqual(1, self.adpt.update_by_path.call_count)
        self.assertEqual(1, mock_lock.call_count)

    @mock.patch('oslo_concurrency.lockutils.lock')
    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridger.'
                '_validate_orphan_on_ensure')
    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridgerVNET.'
                '_find_or_create_vnet')
    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridgerVNET.'
                '_is_arbitrary_vid')
    def test_ensure_vlans_on_nb_new_vlan(self, mock_arb_vid, mock_find_vnet,
                                         mock_orphan_validate, mock_lock):
        """Validates new VLAN on existing Load Group."""
        # Build the responses
        self.adpt.read.side_effect = [self.mgr_nbr_resp, self.mgr_vsw_resp,
                                      self.mgr_vnet_resp]
        mock_arb_vid.return_value = False
        mock_vnet = mock.MagicMock()
        mock_vnet.related_href = 'fake_href'
        mock_find_vnet.return_value = mock_vnet

        def validate_of_update_nb(*kargs, **kwargs):
            # Validate args
            nb = kargs[0]
            self.assertEqual(1, len(nb.load_grps[0].vnet_uri_list))
            self.assertEqual(2, len(nb.load_grps[1].vnet_uri_list))
            self.assertEqual(self.nb_uuid, nb.uuid)
            return nb.entry

        self.adpt.update_by_path.side_effect = validate_of_update_nb

        # Invoke.  VLAN 2227 should be on there already.
        net_br.ensure_vlans_on_nb(self.adpt, self.host_uuid,
                                  self.nb_uuid, [2227, 2000])

        # Validate the calls
        self.assertEqual(1, self.adpt.update_by_path.call_count)
        self.assertEqual(1, mock_lock.call_count)

    @mock.patch('oslo_concurrency.lockutils.lock')
    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridger.'
                '_get_orphan_vlans')
    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridger.'
                '_validate_orphan_on_ensure')
    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridgerVNET.'
                '_find_or_create_vnet')
    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridgerVNET.'
                '_find_available_ld_grp')
    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridgerVNET.'
                '_is_arbitrary_vid')
    def test_ensure_vlan_on_nb_new_lg(
            self, mock_arb_vid, mock_avail_lg, mock_find_vnet,
            mock_orphan_validate, mock_orphan_vlans, mock_lock):
        """Validates new VLAN on new Load Group."""
        # Build the responses
        self.adpt.read.side_effect = [self.mgr_nbr_resp, self.mgr_vsw_resp,
                                      self.mgr_vnet_resp]
        mock_arb_vid.return_value = False
        mock_avail_lg.return_value = None
        mock_orphan_vlans.return_value = []

        # Make the fake virtual networks (the new, then the arb vid)
        mock_vnet = mock.MagicMock()
        mock_vnet.related_href = 'fake_href'
        mock_vnet_avid = mock.MagicMock()
        mock_vnet_avid.related_href = 'fake_avid_href'
        mock_find_vnet.side_effect = [mock_vnet, mock_vnet_avid]

        def validate_of_update_nb(*kargs, **kwargs):
            # Validate args
            nb = kargs[0]
            self.assertIsNotNone(nb)
            self.assertEqual(1, len(nb.load_grps[0].vnet_uri_list))
            self.assertEqual(2, len(nb.load_grps[2].vnet_uri_list))
            self.assertEqual(self.nb_uuid, nb.uuid)
            return nb.entry

        self.adpt.update_by_path.side_effect = validate_of_update_nb

        # Invoke
        net_br.ensure_vlan_on_nb(self.adpt, self.host_uuid, self.nb_uuid, 2000)

        # Validate the calls
        self.assertEqual(1, self.adpt.update_by_path.call_count)
        self.assertEqual(1, mock_lock.call_count)

    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridgerVNET.'
                '_find_or_create_vnet')
    def test_reassign_arbitrary_vid(self, mock_find_vnet):
        nb = self._setup_reassign_arbitrary_vid()

        mock_find_vnet.return_value = mock.MagicMock()
        mock_find_vnet.return_value.related_href = 'other'

        # Make this function return itself.
        def return_self(*kargs, **kwargs):
            return kargs[0].entry

        self.adpt.update_by_path.side_effect = return_self

        bridger = net_br.NetworkBridgerVNET(self.adpt, self.host_uuid)
        bridger._reassign_arbitrary_vid(4094, 4093, nb)

        # Make sure the mocks were called
        self.mock_vsw.assert_called_with(self.adpt, parent_type=pvm_ms.System,
                                         parent_uuid=self.host_uuid,
                                         one_result=True, switch_id=0)
        self.assertEqual(1, mock_find_vnet.call_count)
        self.assertEqual(2, self.adpt.update_by_path.call_count)

    @mock.patch('oslo_concurrency.lockutils.lock')
    def test_remove_vlan_from_nb(self, mock_lock):
        """Happy path testing of the remove VLAN from NB."""
        # Mock Data
        self.adpt.read.return_value = self.mgr_nbr_resp

        def validate_update(*kargs, **kwargs):
            # Make sure the load groups are down to just 1 now.
            nb = kargs[0]
            self.assertEqual(1, len(nb.load_grps))
            return nb.entry

        self.adpt.update_by_path.side_effect = validate_update

        net_br.remove_vlan_from_nb(self.adpt, self.host_uuid, self.nb_uuid,
                                   1000)
        self.assertEqual(1, self.adpt.update_by_path.call_count)
        self.assertEqual(1, mock_lock.call_count)

    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridgerVNET.'
                '_find_vnet_uri_from_lg')
    def test_remove_vlan_from_nb_lb(self, mock_find_vnet):
        """Validates a load balance leaves two total LoadGroups."""
        # Mock Data
        mock_find_vnet.return_value = (
            'https://9.1.2.3:12443/rest/api/uom/ManagedSystem/'
            'c5d782c7-44e4-3086-ad15-b16fb039d63b/VirtualNetwork/'
            'e6c0be9f-b974-35f4-855e-2b7192034fae')
        net_bridge = pvm_net.NetBridge.wrap(self.mgr_nbr_resp)[0]
        net_bridge.load_balance = True

        # Run the remove
        bridger = net_br.NetworkBridgerVNET(self.adpt, self.host_uuid)
        bridger._remove_vlan_from_nb(net_bridge, 1000)

        # Validate that we still have two load groups
        self.assertEqual(2, len(net_bridge.load_grps))
        self.assertEqual(0, len(net_bridge.load_grps[1].vnet_uri_list))

    def test_find_or_create_vnet(self):
        """Validates that a vnet is created (and deleted) as part of find."""
        # Load the data
        vnets = pvm_net.VNet.wrap(self.mgr_vnet_resp)
        vsw = pvm_net.VSwitch.wrap(self.mgr_vsw_resp)[0]

        # Set up the mock create
        resp = pvm_net.VNet.bld(self.adpt, 'FakeName', 4094, vsw.href, True)
        mock_resp = adpt.Response('rm', 'rp', 200, 'reason', {})
        mock_resp.entry = resp.entry
        self.adpt.create.return_value = mock_resp

        # Run the code
        bridger = net_br.NetworkBridgerVNET(self.adpt, self.host_uuid)
        ret_val = bridger._find_or_create_vnet(vnets, 4094, vsw)

        # Equality check
        self.assertEqual(4094, ret_val.vlan)
        self.assertTrue(ret_val.tagged)

        # Make sure the delete was called
        self.assertEqual(1, self.adpt.delete_by_href.call_count)

    def test_find_available_lg(self):
        nb = pvm_net.NetBridge.wrap(self.mgr_nbr_resp)[0]
        bridger = net_br.NetworkBridgerVNET(self.adpt, self.host_uuid)
        lg = bridger._find_available_ld_grp(nb)
        self.assertIsNotNone(lg)

    def test_find_available_lg_load_balance(self):
        """Tests finding the Load Group with load balancing enabled."""
        # Set load balancing to True
        nb = pvm_net.NetBridge.wrap(self.mgr_nbr_fo_resp)[0]
        nb.load_balance = True
        bridger = net_br.NetworkBridgerVNET(self.adpt, self.host_uuid)

        # Even though there is a free VEA, it should come back as None.  This
        # is because there is only one free VEA, but we need to balance across
        # two.
        lg = bridger._find_available_ld_grp(nb)
        self.assertIsNone(lg)

    def test_find_available_min_lg(self):
        nb = mock.MagicMock()

        lg_main = mock.MagicMock()
        lg_first_addl = mock.MagicMock()
        lg_first_addl.vnet_uri_list = ['a', 'b', 'c']
        lg_second_addl = mock.MagicMock()
        lg_second_addl.vnet_uri_list = ['e', 'f']
        nb.load_grps = [lg_main, lg_first_addl, lg_second_addl]

        bridger = net_br.NetworkBridgerVNET(self.adpt, self.host_uuid)
        self.assertEqual(lg_second_addl, bridger._find_available_ld_grp(nb))


class TestNetworkBridgerTA(TestNetworkBridger):
    """General tests for the network bridge super class and the VNet impl."""

    def setUp(self):
        super(TestNetworkBridgerTA, self).setUp()
        self.adptfx.set_traits(fx.LocalPVMTraits)

    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridger.'
                '_get_orphan_vlans')
    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridger.'
                '_validate_orphan_on_ensure')
    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridgerTA.'
                '_reassign_arbitrary_vid')
    @mock.patch('pypowervm.wrappers.network.NetBridge.supports_vlan')
    def test_ensure_vlan_on_nb_reassign(
            self, mock_support_vlan, mock_reassign, mock_orphan_validate,
            mock_orphan_vlans):
        """Validates that after update, we support the VLAN."""
        # Have the response
        self.adpt.read.return_value = self.mgr_nbr_resp

        # First call, say that we don't support the VLAN (which is true).
        # Second call, fake out that we now do.
        # Need pairs, as there are two VLANs we are passing in.
        mock_support_vlan.side_effect = [False, False, True, True]
        mock_orphan_vlans.return_value = []

        # Invoke
        net_br.ensure_vlans_on_nb(self.adpt, self.host_uuid, self.nb_uuid,
                                  ['4093', 4094])
        self.assertEqual(2, self.adpt.read.call_count)
        self.assertEqual(1, mock_reassign.call_count)

        # Should be called re-assigning 4094 (old) to 4092.  Shouldn't be
        # 4093 as that is also an additional VLAN.
        mock_reassign.assert_called_once_with(4094, 4092, mock.ANY)

    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridger.'
                '_validate_orphan_on_ensure')
    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridgerTA.'
                '_is_arbitrary_vid')
    def test_ensure_vlan_on_nb_new_vlan(self, mock_arb_vid,
                                        mock_orphan_validate):
        """Validates new VLAN on existing Trunk Adapter."""
        # Build the responses
        self.adpt.read.side_effect = [self.mgr_nbr_resp, self.mgr_vsw_resp,
                                      self.mgr_vnet_resp]
        mock_arb_vid.return_value = False

        def validate_of_update_nb(*kargs, **kwargs):
            # Validate args
            nb = kargs[0]
            self.assertIsNotNone(nb)
            self.assertEqual(0,
                             len(nb.seas[0].primary_adpt.tagged_vlans))
            self.assertEqual(2,
                             len(nb.seas[0].addl_adpts[0].tagged_vlans))
            self.assertEqual(self.nb_uuid, nb.uuid)

            return nb.entry

        self.adpt.update_by_path.side_effect = validate_of_update_nb

        # Invoke
        net_br.ensure_vlan_on_nb(self.adpt, self.host_uuid, self.nb_uuid, 2000)

        # Validate the calls
        self.assertEqual(1, self.adpt.update_by_path.call_count)

    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridger.'
                '_validate_orphan_on_ensure')
    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridgerTA.'
                '_is_arbitrary_vid')
    def test_ensure_vlans_on_nb_new_vlan(self, mock_arb_vid,
                                         mock_orphan_validate):
        """Validates new VLAN on existing Load Group."""
        # Build the responses
        self.adpt.read.side_effect = [self.mgr_nbr_resp, self.mgr_vsw_resp,
                                      self.mgr_vnet_resp]
        mock_arb_vid.return_value = False

        def validate_of_update_nb(*kargs, **kwargs):
            # Validate args
            nb = kargs[0]
            self.assertEqual(0, len(nb.seas[0].primary_adpt.tagged_vlans))
            self.assertEqual(2, len(nb.seas[0].addl_adpts[0].tagged_vlans))
            self.assertEqual(self.nb_uuid, nb.uuid)
            return nb.entry

        self.adpt.update_by_path.side_effect = validate_of_update_nb

        # Invoke.  VLAN 2227 should be on there already.
        net_br.ensure_vlans_on_nb(self.adpt, self.host_uuid,
                                  self.nb_uuid, [2227, 2000])

        # Validate the calls
        self.assertEqual(1, self.adpt.update_by_path.call_count)

    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridger.'
                '_get_orphan_vlans')
    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridger.'
                '_validate_orphan_on_ensure')
    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridgerTA.'
                '_find_available_trunks')
    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridgerTA.'
                '_is_arbitrary_vid')
    def test_ensure_vlan_on_nb_new_trunk(
            self, mock_arb_vid, mock_avail_trunks, mock_orphan_validate,
            mock_orphan_vlans):
        """Validates new VLAN on new Load Group."""
        # Build the responses
        self.adpt.read.side_effect = [self.mgr_nbr_resp, self.mgr_vsw_resp,
                                      self.mgr_vnet_resp]
        mock_arb_vid.return_value = False
        mock_avail_trunks.return_value = None
        mock_orphan_vlans.return_value = []

        def validate_of_update_nb(*kargs, **kwargs):
            # Validate args
            nb = kargs[0]
            self.assertIsNotNone(nb)
            self.assertEqual(0, len(nb.seas[0].primary_adpt.tagged_vlans))
            self.assertEqual(2, len(nb.seas[0].addl_adpts))
            self.assertEqual(self.nb_uuid, nb.uuid)
            return nb.entry

        self.adpt.update_by_path.side_effect = validate_of_update_nb

        # Invoke
        net_br.ensure_vlan_on_nb(self.adpt, self.host_uuid, self.nb_uuid, 2000)

        # Validate the calls
        self.assertEqual(1, self.adpt.update_by_path.call_count)

    def test_reassign_arbitrary_vid(self):
        nb = self._setup_reassign_arbitrary_vid()

        # Make this function return itself.
        def return_self(*kargs, **kwargs):
            nb_wrap = pvm_net.NetBridge.wrap(kargs[0].entry)
            self.assertEqual(4093,
                             nb_wrap.seas[0].addl_adpts[0].pvid)

            return kargs[0].entry

        self.adpt.update_by_path.side_effect = return_self

        bridger = net_br.NetworkBridgerTA(self.adpt, self.host_uuid)
        bridger._reassign_arbitrary_vid(4094, 4093, nb)

        # Make sure the mocks were called.  Only one update needed.
        self.assertEqual(1, self.adpt.update_by_path.call_count)

    def test_remove_vlan_from_nb(self):
        """Happy path testing of the remove VLAN from NB."""
        # Mock Data
        self.adpt.read.return_value = self.mgr_nbr_resp

        def validate_update(*kargs, **kwargs):
            # Make sure the load groups are down to just 1 now.
            nb = kargs[0]
            self.assertEqual(0, len(nb.seas[0].addl_adpts))
            return nb.entry

        self.adpt.update_by_path.side_effect = validate_update

        net_br.remove_vlan_from_nb(self.adpt, self.host_uuid, self.nb_uuid,
                                   1000)
        self.assertEqual(1, self.adpt.update_by_path.call_count)

    def test_find_available_trunks(self):
        nb = pvm_net.NetBridge.wrap(self.mgr_nbr_resp)[0]
        bridger = net_br.NetworkBridgerTA(self.adpt, self.host_uuid)
        trunks = bridger._find_available_trunks(nb)
        self.assertIsNotNone(trunks)

    def test_find_available_trunks_load_balance(self):
        """Tests finding the trunk with load balancing enabled."""
        # Set load balancing to True
        nb = pvm_net.NetBridge.wrap(self.mgr_nbr_fo_resp)[0]
        nb.load_balance = True
        bridger = net_br.NetworkBridgerTA(self.adpt, self.host_uuid)

        # Even though there is a free VEA, it should come back as None.  This
        # is because there is only one free VEA, but we need to balance across
        # two.
        trunks = bridger._find_available_trunks(nb)
        self.assertIsNone(trunks)

    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridgerTA._trunk_list')
    def test_find_available_min_trunk(self, mock_trunk_list):
        nb = mock.MagicMock()

        trunk_addl = mock.MagicMock()
        trunk_addl.tagged_vlans = ['a', 'b', 'c']
        trunk_addl2 = mock.MagicMock()
        trunk_addl2.tagged_vlans = ['e', 'f']
        trunk_addl3 = mock.MagicMock()
        trunk_addl3.tagged_vlans = ['g', 'h', 'i']

        sea = mock.MagicMock()
        sea.addl_adpts = [trunk_addl, trunk_addl2, trunk_addl3]
        nb.seas = [sea]

        bridger = net_br.NetworkBridgerTA(self.adpt, self.host_uuid)
        bridger._find_available_trunks(nb)

        # Validate the trunk list is called with the second additional adapter
        mock_trunk_list.assert_called_with(nb, trunk_addl2)

    def test_find_peer_trunk(self):
        bridger = net_br.NetworkBridgerTA(self.adpt, self.host_uuid)

        # No failover, shouldn't have a peer
        nbs = pvm_net.NetBridge.wrap(self.mgr_nbr_resp)
        resp = bridger._find_peer_trunk(nbs[0], nbs[0].seas[0].primary_adpt)
        self.assertIsNone(resp)

        # Failover, should have a peer
        nbs = pvm_net.NetBridge.wrap(self.mgr_nbr_fo_resp)
        resp = bridger._find_peer_trunk(nbs[0], nbs[0].seas[0].primary_adpt)
        self.assertEqual(nbs[0].seas[1].primary_adpt, resp)

    @mock.patch('pypowervm.tasks.network_bridger.NetworkBridgerTA.'
                '_reassign_arbitrary_vid')
    def test_remove_vlan_from_nb_arb_vid(self, mock_reassign):
        """Attempt to remove an arbitrary VID off the network bridge."""
        # Mock Data
        self.adpt.read.return_value = self.mgr_nbr_fo_resp

        # Run the remove of the VLAN.  Make sure it is invoked with a new
        # valid arbitrary PVID.
        net_br.remove_vlan_from_nb(self.adpt, self.host_uuid, self.nb_uuid,
                                   '4094')
        self.assertEqual(1, mock_reassign.call_count)
        mock_reassign.assert_called_with(4094, 4093, mock.ANY)

    def test_remove_vlan_from_nb_lb(self):
        """Validates a load balance remove leaves an additional adpt."""
        # Mock Data
        net_bridge = pvm_net.NetBridge.wrap(self.mgr_nbr_resp)[0]
        net_bridge.load_balance = True

        # Run the remove
        bridger = net_br.NetworkBridgerTA(self.adpt, self.host_uuid)
        bridger._remove_vlan_from_nb(net_bridge, 1000)

        # Validate that we left the trunk but no new additional VLANs
        self.assertEqual(1, len(net_bridge.seas[0].addl_adpts))
        self.assertEqual(0, len(net_bridge.seas[0].addl_adpts[0].tagged_vlans))



# Copyright 2014, 2017 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Wrappers for virtual storage elements and adapters."""

import abc
import base64
import binascii
from oslo_log import log as logging
import six

import pypowervm.const as c
import pypowervm.exceptions as ex
from pypowervm.i18n import _
import pypowervm.util as u
import pypowervm.wrappers.entry_wrapper as ewrap

LOG = logging.getLogger(__name__)

UDID = 'UniqueDeviceID'

# Virtual Disk Constants
DISK_ROOT = 'VirtualDisk'
_DISK_CAPACITY = 'DiskCapacity'
_DISK_LABEL = 'DiskLabel'
DISK_NAME = 'DiskName'
_DISK_MAX_LOGICAL_VOLS = 'MaxLogicalVolumes'
_DISK_PART_SIZE = 'PartitionSize'
_DISK_VG = 'VolumeGroup'
_DISK_BASE = 'BaseImage'
_DISK_UDID = UDID
_DISK_TYPE = 'VirtualDiskType'
_DISK_BACKSTORE_TYPE = 'BackStoreType'
_VDISK_EL_ORDER = [_DISK_CAPACITY, _DISK_LABEL, DISK_NAME,
                   _DISK_MAX_LOGICAL_VOLS, _DISK_PART_SIZE, _DISK_VG,
                   _DISK_BASE, _DISK_UDID, _DISK_TYPE, _DISK_BACKSTORE_TYPE]


class VDiskType(object):
    """From VirtualDiskType.Enum."""
    FILE = 'File'
    LV = 'LogicalVolume'


class BackStoreType(object):
    """From BackStoreType.Enum."""
    # A kernel-space handler that supports raw files.
    FILE_IO = 'fileio'
    # A user-space handler that supports RAW, QCOW or QCOW2 files.
    USER_QCOW = 'user:qcow'

# Physical Volume Constants
PVS = 'PhysicalVolumes'
PHYS_VOL = 'PhysicalVolume'
_PV_AVAIL_PHYS_PART = 'AvailablePhysicalPartitions'
_PV_VOL_DESC = 'Description'
_PV_LOC_CODE = 'LocationCode'
_PV_PERSISTENT_RESERVE = 'PersistentReserveKeyValue'
_PV_RES_POLICY = 'ReservePolicy'
_PV_RES_POLICY_ALGO = 'ReservePolicyAlgorithm'
_PV_TOTAL_PHYS_PARTS = 'TotalPhysicalPartitions'
_PV_UDID = UDID
_PV_AVAIL_FOR_USE = 'AvailableForUsage'
_PV_VOL_SIZE = 'VolumeCapacity'
_PV_VOL_NAME = 'VolumeName'
_PV_VOL_STATE = 'VolumeState'
_PV_VOL_UNIQUE_ID = 'VolumeUniqueID'
_PV_FC_BACKED = 'IsFibreChannelBacked'
_PV_STG_LABEL = 'StorageLabel'
_PV_PG83 = 'DescriptorPage83'
_PV_EL_ORDER = [_PV_AVAIL_PHYS_PART, _PV_VOL_DESC, _PV_LOC_CODE,
                _PV_PERSISTENT_RESERVE, _PV_RES_POLICY, _PV_RES_POLICY_ALGO,
                _PV_TOTAL_PHYS_PARTS, _PV_UDID, _PV_AVAIL_FOR_USE,
                _PV_VOL_SIZE, _PV_VOL_NAME, _PV_VOL_STATE, _PV_VOL_UNIQUE_ID,
                _PV_FC_BACKED, _PV_STG_LABEL, _PV_PG83]

# Virtual Optical Media Constants
VOPT_ROOT = 'VirtualOpticalMedia'
VOPT_NAME = 'MediaName'
_VOPT_SIZE = 'Size'
_VOPT_UDID = 'MediaUDID'
_VOPT_MOUNT_TYPE = 'MountType'
_VOPT_EL_ORDER = [VOPT_NAME, _VOPT_UDID, _VOPT_MOUNT_TYPE, _VOPT_SIZE]

# Virtual Media Repository Constants
_VREPO_ROOT = 'VirtualMediaRepository'
_VREPO_OPTICAL_MEDIA_ROOT = 'OpticalMedia'
_VREPO_NAME = 'RepositoryName'
_VREPO_SIZE = 'RepositorySize'
_VREPO_EL_ORDER = [_VREPO_OPTICAL_MEDIA_ROOT, _VREPO_NAME, _VREPO_SIZE]

# Volume Group Constants
_VG_AVAILABLE_SIZE = 'AvailableSize'
_VG_BACKING_DEVICE_COUNT = 'BackingDeviceCount'
_VG_FREE_SPACE = 'FreeSpace'
_VG_CAPACITY = 'GroupCapacity'
_VG_NAME = 'GroupName'
_VG_SERIAL_ID = 'GroupSerialID'
_VG_STATE = 'GroupState'
_VG_MAX_LVS = 'MaximumLogicalVolumes'
_VG_MEDIA_REPOS = 'MediaRepositories'
_VG_MIN_ALLOC_SIZE = 'MinimumAllocationSize'
_VG_PHS_VOLS = PVS
_VG_UDID = UDID
_VG_VDISKS = 'VirtualDisks'
_VG_EL_ORDER = (_VG_AVAILABLE_SIZE, _VG_BACKING_DEVICE_COUNT, _VG_FREE_SPACE,
                _VG_CAPACITY, _VG_NAME, _VG_SERIAL_ID, _VG_STATE, _VG_MAX_LVS,
                _VG_MEDIA_REPOS, _VG_MIN_ALLOC_SIZE, _VG_PHS_VOLS, _VG_UDID,
                _VG_VDISKS)

# LogicalUnit Constants
_LU_THIN = 'ThinDevice'
_LU_UDID = UDID
_LU_CAPACITY = 'UnitCapacity'
_LU_TYPE = 'LogicalUnitType'
_LU_CLONED_FROM = 'ClonedFrom'
_LU_IN_USE = 'InUse'
_LU_NAME = 'UnitName'
_LU_EL_ORDER = (_LU_THIN, _LU_UDID, _LU_CAPACITY, _LU_TYPE, _LU_CLONED_FROM,
                _LU_IN_USE, _LU_NAME)


class LUType(object):
    DISK = "VirtualIO_Disk"
    HIBERNATION = "VirtualIO_Hibernation"
    IMAGE = "VirtualIO_Image"
    AMS = "VirtualIO_Active_Memory_Sharing"

_CAPACITY = 'Capacity'

# Tier Constants
_TIER_NAME = 'Name'
_TIER_UDID = UDID
_TIER_IS_DEFAULT = 'IsDefault'
_TIER_CAPACITY = _CAPACITY
_TIER_ASSOC_SSP = 'AssociatedSharedStoragePool'

# Shared Storage Pool Constants
_SSP_NAME = 'StoragePoolName'
_SSP_UDID = UDID
_SSP_CAPACITY = _CAPACITY
_SSP_FREE_SPACE = 'FreeSpace'
_SSP_TOTAL_LU_SIZE = 'TotalLogicalUnitSize'
_SSP_LUS = 'LogicalUnits'
_SSP_LU = 'LogicalUnit'
_SSP_OCS = 'OverCommitSpace'
_SSP_PVS = PVS
_SSP_PV = PHYS_VOL

# Virtual Adapter Constants
CLIENT_ADPT = 'ClientAdapter'
SERVER_ADPT = 'ServerAdapter'

# Common to all Virtual Adapters
_VADPT_TYPE = 'AdapterType'
_VADPT_DRC_NAME = 'DynamicReconfigurationConnectorName'
_VADPT_LOC_CODE = 'LocationCode'
_VADPT_LOCAL_ID = 'LocalPartitionID'
_VADPT_REQD = 'RequiredAdapter'
_VADPT_VARIED_ON = 'VariedOn'
_VADPT_NEXT_SLOT = 'UseNextAvailableSlotID'
_VADPT_NEXT_HI_SLOT = 'UseNextAvailableHighSlotID'
_VADPT_SLOT_NUM = 'VirtualSlotNumber'
_VADPT_ENABLED = 'Enabled'
_VADPT_NAME = 'AdapterName'
_VADPT_UDID = 'UniqueDeviceID'

# Common to VSCSI Adapters (Client & Server)
_VSCSI_ADPT_BACK_DEV_NAME = 'BackingDeviceName'
_VSCSI_ADPT_REM_BACK_DEV_NAME = 'RemoteBackingDeviceName'
_VSCSI_ADPT_REM_LPAR_ID = 'RemoteLogicalPartitionID'
_VSCSI_ADPT_REM_SLOT_NUM = 'RemoteSlotNumber'
_VSCSI_ADPT_SVR_LOC_CODE = 'ServerLocationCode'

# Common to Client Adapters
_VCLNT_ADPT_SVR_ADPT = SERVER_ADPT

# Common to VFC Adapters (Client & Server)
_VFC_ADPT_CONN_PARTITION = 'ConnectingPartition'
_VFC_ADPT_CONN_PARTITION_ID = 'ConnectingPartitionID'
_VFC_ADPT_CONN_SLOT_NUM = 'ConnectingVirtualSlotNumber'

# VFC Server Adapter-specific
_VFC_SVR_ADPT_MAP_PORT = 'MapPort'
_VFC_SVR_ADPT_PHYS_PORT = 'PhysicalPort'

# VFC Client Adapter-specific
_VFC_CLNT_ADPT_WWPNS = 'WWPNs'
_VFC_CLNT_ADPT_LOGGED_IN = 'NportLoggedInStatus'
_VFC_CLNT_ADPT_OS_DISKS = 'OperatingSystemDisks'

# Element Ordering:
#
# A <ServerAdapter/> might be a VSCSI server adapter or a VFC server adapter.
# Likewise <ClientAdapter/>.  The schema inheritance hierarchy informs the
# way we build up the element order constants:
#
#                            VirtualIOAdapter
#                   VFCAdapter             VSCSIAdapter == VSCSIServerAdapter
#      VFCClientAdapter  VFCServerAdapter        VSCSIClientAdapter
#
# However, this doesn't match up with the hierarchy of our wrapper classes:
#
#               VClientStorageAdapterElement
# VSCSIClientAdapterElement    VFCClientAdapterElement
#
#               VServerStorageAdapterElement
# VSCSIServerAdapterElement    VFCServerAdapterElement
#
# So we have to get creative with element ordering for the base classes, since
# they hold the @pvm_type decorator.  We interleave the VSCSI and VFC
# properties to create an element order that can be used commonly for both
# types.  This only works because all overlapping properties happen to be in
# the same order.
#
# Yes, this is funky.

# Converged ordering base for VFC and VSCSI adapters
_VADPT_BASE_EL_ORDER = (
    _VADPT_TYPE, _VADPT_DRC_NAME, _VADPT_LOC_CODE, _VADPT_LOCAL_ID,
    _VADPT_REQD, _VADPT_VARIED_ON, _VADPT_NEXT_SLOT, _VADPT_NEXT_HI_SLOT,
    _VADPT_SLOT_NUM, _VADPT_ENABLED, _VADPT_NAME, _VSCSI_ADPT_BACK_DEV_NAME,
    _VSCSI_ADPT_REM_BACK_DEV_NAME, _VSCSI_ADPT_REM_LPAR_ID,
    _VFC_ADPT_CONN_PARTITION, _VFC_ADPT_CONN_PARTITION_ID,
    _VSCSI_ADPT_REM_SLOT_NUM, _VFC_ADPT_CONN_SLOT_NUM,
    _VSCSI_ADPT_SVR_LOC_CODE, _VADPT_UDID)

# Converged (VSCSI & VFC) Server Adapter element order
_V_SVR_ADPT_EL_ORDER = _VADPT_BASE_EL_ORDER + (
    _VFC_SVR_ADPT_MAP_PORT, _VFC_SVR_ADPT_PHYS_PORT)

# Converged (VSCSI & VFC) Client Adapter element order
_V_CLNT_ADPT_EL_ORDER = _VADPT_BASE_EL_ORDER + (
    _VCLNT_ADPT_SVR_ADPT, _VFC_CLNT_ADPT_WWPNS, _VFC_CLNT_ADPT_LOGGED_IN,
    _VFC_CLNT_ADPT_OS_DISKS)

VFC_CLIENT_ADPT = 'VirtualFibreChannelClientAdapter'

# TargetDevice Constants
_TD_LU_TD = 'SharedStoragePoolLogicalUnitVirtualTargetDevice'
_TD_PV_TD = 'PhysicalVolumeVirtualTargetDevice'
_TD_VOPT_TD = 'VirtualOpticalTargetDevice'
_TD_VDISK_TD = 'LogicalVolumeVirtualTargetDevice'
_TD_LUA = 'LogicalUnitAddress'
_TD_NAME = 'TargetName'


@ewrap.EntryWrapper.pvm_type('VolumeGroup', child_order=_VG_EL_ORDER)
class VG(ewrap.EntryWrapper):
    """Represents a Volume Group that resides on the Virtual I/O Server."""

    @classmethod
    def bld(cls, adapter, name, pv_list):
        vg = super(VG, cls)._bld(adapter)
        vg.name = name
        vg.phys_vols = pv_list
        return vg

    @property
    def name(self):
        return self._get_val_str(_VG_NAME)

    @name.setter
    def name(self, val):
        self.set_parm_value(_VG_NAME, val)

    @property
    def capacity(self):
        """Overall capacity in GB (float)."""
        return self._get_val_float(_VG_CAPACITY)

    @property
    def available_size(self):
        """Available size for new volumes in GB (float)."""
        return self._get_val_float(_VG_AVAILABLE_SIZE)

    @property
    def free_space(self):
        """Current free space in GB (float)."""
        return self._get_val_float(_VG_FREE_SPACE)

    @property
    def serial_id(self):
        return self._get_val_str(_VG_SERIAL_ID)

    @property
    def vmedia_repos(self):
        """Returns a list of  wrappers."""
        es = ewrap.WrapperElemList(self._find_or_seed(_VG_MEDIA_REPOS),
                                   VMediaRepos)
        return es

    @vmedia_repos.setter
    def vmedia_repos(self, repos):
        """Replaces the VirtualMediaRepositories with the new value.

        :param repos: A list of VMediaRepos objects that will
                      replace the existing repositories.
        """
        self.replace_list(_VG_MEDIA_REPOS, repos)

    @property
    def phys_vols(self):
        """Returns a list of the Physical Volumes that back this repo."""
        # TODO(efried): parent_entry=self not needed once VIOS supports pg83
        # descriptor in Events
        es = ewrap.WrapperElemList(self._find_or_seed(_VG_PHS_VOLS), PV,
                                   parent_entry=self)
        return es

    @phys_vols.setter
    def phys_vols(self, phys_vols):
        """Replaces the physical volumes with the new value.

        :param phys_vols: A list of PV objects that will replace
                          the existing Physcial Volumes.
        """
        self.replace_list(_VG_PHS_VOLS, phys_vols)

    @property
    def virtual_disks(self):
        """Returns a list of the Virtual Disks that are in the repo."""
        es = ewrap.WrapperElemList(self._find_or_seed(_VG_VDISKS), VDisk)
        return es

    @virtual_disks.setter
    def virtual_disks(self, virt_disks):
        """Replaces the virtual disks with the new value.

        :param virt_disks: A list of VDisk objects that will replace
                           the existing Virtual Disks.
        """
        self.replace_list(_VG_VDISKS, virt_disks)


@ewrap.ElementWrapper.pvm_type(_VREPO_ROOT, has_metadata=True,
                               child_order=_VREPO_EL_ORDER)
class VMediaRepos(ewrap.ElementWrapper):
    """A Virtual Media Repository for a VIOS.

    Typically used to store an ISO file for image building.
    """

    @classmethod
    def bld(cls, adapter, name, size):
        """Creates a fresh VMediaRepos wrapper.

        This should be used when adding a new Virtual Media Repository to a
        Volume Group.  The name and size for the media repository is required.
        The other attributes are generated from the system.

        Additionally, once created, specific VirtualOpticalMedia can be added
        onto the object.

        :param adapter: A pypowervm.adapter.Adapter (for traits, etc.)
        :param name: The name of the Virtual Media Repository.
        :param size: The size of the repository in GB (float).
        :returns: A VMediaRepos wrapper that can be used for create.
        """
        vmr = super(VMediaRepos, cls)._bld(adapter)
        vmr._name(name)
        vmr._size(size)
        return vmr

    @property
    def optical_media(self):
        """Returns a list of the VirtualOpticalMedia devices in the repo."""
        seed = self._find_or_seed(_VREPO_OPTICAL_MEDIA_ROOT)
        return ewrap.WrapperElemList(seed, VOptMedia)

    @optical_media.setter
    def optical_media(self, new_media):
        """Sets the list of VirtualOpticalMedia devices in the repo.

        :param new_media: The list of new VOptMedia.
        """
        self.replace_list(_VREPO_OPTICAL_MEDIA_ROOT, new_media)

    @property
    def name(self):
        return self._get_val_str(_VREPO_NAME)

    def _name(self, new_name):
        self.set_parm_value(_VREPO_NAME, new_name)

    @property
    def size(self):
        """Returns the size in GB (float)."""
        return self._get_val_float(_VREPO_SIZE)

    def _size(self, new_size):
        self.set_float_gb_value(_VREPO_SIZE, new_size)


@six.add_metaclass(abc.ABCMeta)
@ewrap.Wrapper.base_pvm_type
class _VTargetDevMethods(ewrap.Wrapper):
    """Base class for {storage_type}TargetDevice of an active VSCSIMapping."""

    @classmethod
    def bld(cls, adapter, lua=None, name=None):
        """Build a new Virtual Target Device.

        :param adapter: A pypowervm.adapter.Adapter (for traits, etc.)
        :param lua: (Optional, Default None) Logical Unit Address string to
                    assign to the new VTD.
        :param name: (Optional, Default None) Name of the TargetDev. If None
                     name will be assigned by the server
        :return: A new {storage_type}TargetDev, where {storage_type} is
                 appropriate to the subclass.
        """
        vtd = super(_VTargetDevMethods, cls)._bld(adapter)
        if lua is not None:
            vtd._lua(lua)
        if name is not None:
            vtd._name(name)
        return vtd

    @property
    def lua(self):
        """Logical Unit Address of the target device."""
        return self._get_val_str(_TD_LUA)

    def _lua(self, val):
        """Set the Logical Unit Address of this target device."""
        self.set_parm_value(_TD_LUA, val)

    @property
    def name(self):
        """Target Name of the device"""
        return self._get_val_str(_TD_NAME)

    def _name(self, val):
        """Set the Target Name of the device"""
        self.set_parm_value(_TD_NAME, val)


@ewrap.ElementWrapper.pvm_type(_TD_LU_TD, has_metadata=True)
class LUTargetDev(_VTargetDevMethods, ewrap.ElementWrapper):
    """SSP Logical Unit Virtual Target Device for a VSCSIMapping."""
    pass


@ewrap.ElementWrapper.pvm_type(_TD_PV_TD, has_metadata=True)
class PVTargetDev(_VTargetDevMethods, ewrap.ElementWrapper):
    """Physical Volume Virtual Target Device for a VSCSIMapping."""
    pass


@ewrap.ElementWrapper.pvm_type(_TD_VDISK_TD, has_metadata=True)
class VDiskTargetDev(_VTargetDevMethods, ewrap.ElementWrapper):
    """Virtual Disk (Logical Volume) Target Device for a VSCSIMapping."""
    pass


@ewrap.ElementWrapper.pvm_type(_TD_VOPT_TD, has_metadata=True)
class VOptTargetDev(_VTargetDevMethods, ewrap.ElementWrapper):
    """Virtual Optical Media Target Device for a VSCSIMapping."""
    pass


@ewrap.ElementWrapper.pvm_type(VOPT_ROOT, has_metadata=True,
                               child_order=_VOPT_EL_ORDER)
class VOptMedia(ewrap.ElementWrapper):
    """A virtual optical piece of media."""
    target_dev_type = VOptTargetDev

    @classmethod
    def bld(cls, adapter, name, size=None, mount_type='rw'):
        """Creates a fresh VOptMedia wrapper.

        This should be used when adding a new VirtualOpticalMedia device to a
        VirtualMediaRepository.

        :param adapter: A pypowervm.adapter.Adapter (for traits, etc.)
        :param name: The device name.
        :param size: The device size in GB, decimal precision.
        :param mount_type: The type of mount.  Defaults to RW.  Can be set to R
        :returns: A VOptMedia wrapper that can be used for create.
        """
        vom = super(VOptMedia, cls)._bld(adapter)
        vom._media_name(name)
        if size is not None:
            vom._size(size)
        vom._mount_type(mount_type)
        return vom

    @classmethod
    def bld_ref(cls, adapter, name):
        """Creates a VOptMedia wrapper for referencing an existing VOpt."""
        vom = super(VOptMedia, cls)._bld(adapter)
        vom._media_name(name)
        return vom

    @property
    def media_name(self):
        return self._get_val_str(VOPT_NAME)

    @property
    def name(self):
        """Same as media_name - for consistency with other storage types."""
        return self.media_name

    def _media_name(self, new_name):
        self.set_parm_value(VOPT_NAME, new_name)

    @property
    def size(self):
        """Size is a float represented in GB."""
        return self._get_val_float(_VOPT_SIZE)

    def _size(self, new_size):
        self.set_float_gb_value(_VOPT_SIZE, new_size)

    @property
    def udid(self):
        return self._get_val_str(_VOPT_UDID)

    @property
    def mount_type(self):
        return self._get_val_str(_VOPT_MOUNT_TYPE)

    def _mount_type(self, new_mount_type):
        self.set_parm_value(_VOPT_MOUNT_TYPE, new_mount_type)


@ewrap.ElementWrapper.pvm_type(PHYS_VOL, has_metadata=True,
                               child_order=_PV_EL_ORDER)
class PV(ewrap.ElementWrapper):
    """A physical volume that backs a Volume Group."""
    target_dev_type = PVTargetDev

    @classmethod
    def bld(cls, adapter, name, udid=None):
        """Creates the a fresh PV wrapper.

        This should be used when wishing to add physical volumes to a Volume
        Group.  Only the name is required.  The other attributes are generated
        from the system.

        The name matches the device name on the system.

        :param adapter: A pypowervm.adapter.Adapter (for traits, etc.)
        :param name: The name of the physical volume on the Virtual I/O Server
                     to add to the Volume Group.  Ex. 'hdisk1'.
        :param udid: Universal Disk Identifier.
        :returns: An Element that can be used for a PhysicalVolume create.
        """
        pv = super(PV, cls)._bld(adapter)
        # Assignment order is significant
        if udid:
            pv.udid = udid
        pv.name = name
        return pv

    @property
    def udid(self):
        """The unique device id."""
        return self._get_val_str(_PV_UDID)

    @udid.setter
    def udid(self, new_udid):
        self.set_parm_value(_PV_UDID, new_udid)

    @property
    def capacity(self):
        """Returns the capacity as an int in MB."""
        return self._get_val_int(_PV_VOL_SIZE)

    @property
    def name(self):
        return self._get_val_str(_PV_VOL_NAME)

    @name.setter
    def name(self, newname):
        self.set_parm_value(_PV_VOL_NAME, newname)

    @property
    def state(self):
        return self._get_val_str(_PV_VOL_STATE)

    @property
    def is_fc_backed(self):
        return self._get_val_bool(_PV_FC_BACKED)

    @property
    def description(self):
        return self._get_val_str(_PV_VOL_DESC)

    @property
    def loc_code(self):
        return self._get_val_str(_PV_LOC_CODE)

    @property
    def avail_for_use(self):
        return self._get_val_bool(_PV_AVAIL_FOR_USE)

    @property
    def pg83(self):
        encoded = self._get_val_str(_PV_PG83)
        # TODO(efried): Temporary workaround until VIOS supports pg83 in Events
        # >>>CUT HERE>>>
        if not encoded:
            # The PhysicalVolume XML doesn't contain the DescriptorPage83
            # property.  (This could be because the disk really doesn't have
            # this attribute; but if the caller is asking for pg83, they likely
            # expect that it should.)  More likely, it is because their VIOS is
            # running at a level which supplies this datum in a fresh inventory
            # query, but not in a PV ADD Event.  In that case, use the
            # LUARecovery Job to perform the fresh inventory query to retrieve
            # this value.  Since this is expensive, we cache the value.
            if not hasattr(self, '_pg83_encoded'):
                # Get the VIOS UUID from the parent_entry of this PV.  Raise if
                # it doesn't exist.
                if not hasattr(self, 'parent_entry') or not self.parent_entry:
                    raise ex.UnableToBuildPG83EncodingMissingParent(
                        dev_name=self.name)
                # The parent_entry is either a VG or a VIOS.  If a VG, it is a
                # child of the owning VIOS, so pull out the ROOT UUID of its
                # href. If a VIOS, we can't count on the href being a root URI,
                # so pull the target UUID regardless.
                use_root_uuid = isinstance(self.parent_entry, VG)
                vio_uuid = u.get_req_path_uuid(
                    self.parent_entry.href, preserve_case=True,
                    root=use_root_uuid)

                # Local import to prevent circular dependency
                from pypowervm.tasks import hdisk
                # Cache the encoded value for performance
                self._pg83_encoded = hdisk.get_pg83_via_job(
                    self.adapter, vio_uuid, self.udid)
            encoded = self._pg83_encoded
        # <<<CUT HERE<<<
        try:
            return base64.b64decode(encoded).decode(
                'utf-8') if encoded else None
        except (TypeError, binascii.Error) as te:
            LOG.warning(_('PV had encoded pg83 descriptor "%(pg83_raw)s", but '
                          'it failed to decode (%(type_error)s).'),
                        {'pg83_raw': encoded, 'type_error': te.args[0]})
        return None


@ewrap.Wrapper.base_pvm_type
class _VDisk(ewrap.ElementWrapper):
    """Methods common to VDisk and FileIO."""

    @property
    def name(self):
        return self._get_val_str(DISK_NAME)

    @name.setter
    def name(self, name):
        self.set_parm_value(DISK_NAME, name)

    @property
    def label(self):
        return self._get_val_str(_DISK_LABEL)

    def _label(self, new_label):
        self.set_parm_value(_DISK_LABEL, new_label)

    @property
    def capacity(self):
        """Returns the capacity in GB (float)."""
        return self._get_val_float(_DISK_CAPACITY)

    @capacity.setter
    def capacity(self, capacity):
        self.set_float_gb_value(_DISK_CAPACITY, capacity)

    @property
    def udid(self):
        return self._get_val_str(_DISK_UDID)

    @property
    def vdtype(self):
        return self._get_val_str(_DISK_TYPE)

    def _vdtype(self, val):
        self.set_parm_value(_DISK_TYPE, val, attrib=c.ATTR_KSV150)

    def _base_image(self, base_image):
        self.set_parm_value(_DISK_BASE, base_image)

    @property
    def backstore_type(self):
        """The backing store type, one of the BackStoreType enum values."""
        return self._get_val_str(_DISK_BACKSTORE_TYPE)

    def _backstore_type(self, val):
        """Set the backing store type.

        :param val: One of the BackStoreType enum values.
        """
        self.set_parm_value(_DISK_BACKSTORE_TYPE, val, attrib=c.ATTR_KSV150)


@ewrap.ElementWrapper.pvm_type(DISK_ROOT, has_metadata=True,
                               child_order=_VDISK_EL_ORDER)
class FileIO(_VDisk):
    """A special case of VirtualDisk representing a File I/O object.

    Do not PUT (.create) this wrapper directly.  Attach it to a VSCSIMapping
    and PUT that instead.
    """
    target_dev_type = VDiskTargetDev

    @classmethod
    def bld_ref(cls, adapter, path, backstore_type=None):
        """Creates a FileIO reference for inclusion in a VSCSIMapping.

        :param adapter: A pypowervm.adapter.Adapter for the REST API.
        :param path: The file system path of the File I/O object.
        :return: An Element that can be attached to a VSCSIMapping to create a
                 File I/O mapping on the server.
        """
        fio = super(FileIO, cls)._bld(adapter)
        fio._label(path)
        fio.name = path
        fio._vdtype(VDiskType.FILE)
        if backstore_type is not None:
            fio._backstore_type(backstore_type)
        return fio

    # Maintained for backward compatibility.  FileIOs aren't created by REST.
    bld = bld_ref

    @property
    def path(self):
        """Alias for 'label'."""
        return self.label


@ewrap.ElementWrapper.pvm_type(DISK_ROOT, has_metadata=True,
                               child_order=_VDISK_EL_ORDER)
class VDisk(_VDisk):
    """A virtual disk that can be attached to a VM."""
    target_dev_type = VDiskTargetDev

    @classmethod
    def bld(cls, adapter, name, capacity, label=None, base_image=None):
        """Creates a VDisk Wrapper for creating a new VDisk.

        This should be used when the user wishes to add a new Virtual Disk to
        the Volume Group.  The flow is to use this method to lay out the
        attributes of the new Virtual Disk.  Then add it to the Volume Group's
        virtual disks. Then perform an update of the Volume Group.  The disk
        should be created by the update operation.

        :param adapter: A pypowervm.adapter.Adapter (for traits, etc.)
        :param name: The name of the virtual disk
        :param capacity: A float number that defines the GB of the disk.
        :param label: The generic label for the disk.  Not required.
        :param base_image: UDID of virtual disk that contains source data
                           Not required.
        :returns: An Element that can be used for a VirtualDisk create.
        """
        vd = super(VDisk, cls)._bld(adapter)
        vd.capacity = capacity
        # Label must be specified; str will make None 'None'.
        vd._label(str(label))
        vd.name = name
        if base_image:
            vd._base_image(base_image)
        return vd

    @classmethod
    def bld_ref(cls, adapter, name):
        """Creates a VDisk Wrapper for referring to an existing VDisk."""
        vd = super(VDisk, cls)._bld(adapter)
        vd.name = name
        return vd

    @property
    def vg_uri(self):
        return self.get_href(_DISK_VG, one_result=True)


@six.add_metaclass(abc.ABCMeta)
@ewrap.Wrapper.base_pvm_type
class _LUBase(ewrap.Wrapper):
    """Mixin for a Logical Unit EntryWrapper or ElementWrapper.

    A Logical Unit is either a DETAIL object (within a SharedStoragePool or
    SCSI mapping); or it is a first-class REST CHILD of Tier.  In either case,
    its properties/methods are the same, provided here.
    """
    target_dev_type = LUTargetDev

    @classmethod
    def bld(cls, adapter, name, capacity, thin=None, typ=None, clone=None):
        """Build a fresh wrapper for LU creation within an SSP.

        :param adapter: A pypowervm.adapter.Adapter (for traits, etc.)
        :param name: The name to assign to the new LogicalUnit
        :param capacity: Capacity in GB for the new LogicalUnit
        :param thin: Provision the new LU as thin (True) or thick (False).
        :param typ: Logical Unit type, one of the LUType values.
        :param clone: If the new LU is to be a linked clone, this param is a
                      LU(Ent) wrapper representing the backing image LU.
        :return: A new LU wrapper suitable for adding to SSP.logical_units
                 prior to update.
        """
        lu = super(_LUBase, cls)._bld(adapter)
        lu._name(name)
        lu._capacity(capacity)
        if thin is not None:
            lu._is_thin(thin)
        if typ is not None:
            lu._lu_type(typ)
        if clone is not None:
            lu._cloned_from_udid(clone.udid)
            # New LU must be at least as big as the backing LU.
            lu._capacity(max(capacity, clone.capacity))
        return lu

    @classmethod
    def bld_ref(cls, adapter, name, udid):
        """Creates the a fresh LU wrapper.

        The name matches the device name on the system.

        :param adapter: A pypowervm.adapter.Adapter (for traits, etc.)
        :param name: The name of the logical unit on the Virtual I/O Server.
        :param udid: Universal Disk Identifier.
        :returns: An Element that can be used for a PhysicalVolume create.
        """
        lu = super(_LUBase, cls)._bld(adapter)
        lu._name(name)
        lu._udid(udid)
        return lu

    def __eq__(self, other):
        """Name and UDID are sufficient for equality.

        For example, if we change an LU's capacity, it's still the same LU.
        We're counting on UDIDs not being repeated in any reasonable scenario.
        """
        return self.name == other.name and self.udid == other.udid

    def __hash__(self):
        """For comparing sets of LUs."""
        # The contract of hash is that two equal thingies must have the same
        # hash, but two thingies with the same hash are not necessarily equal.
        # The hash is used for assigning keys to hash buckets in a dictionary:
        # if two keys hash the same, their items go into the same bucket, but
        # they're still different items.
        if six.PY3:
            conv = int
        else:
            import __builtin__
            conv = __builtin__.long
        return conv(self.udid[2:], base=16)

    @property
    def name(self):
        return self._get_val_str(_LU_NAME)

    def _name(self, value):
        return self.set_parm_value(_LU_NAME, value)

    @property
    def udid(self):
        return self._get_val_str(_LU_UDID)

    def _udid(self, value):
        self.set_parm_value(_LU_UDID, value)

    @property
    def capacity(self):
        """Float capacity in GB."""
        return self._get_val_float(_LU_CAPACITY)

    def _capacity(self, val):
        """val is float."""
        self.set_float_gb_value(_LU_CAPACITY, val)

    @property
    def lu_type(self):
        """String enum value e.g. "VirtualIO_Disk."""
        return self._get_val_str(_LU_TYPE)

    def _lu_type(self, val):
        self.set_parm_value(_LU_TYPE, val)

    @property
    def is_thin(self):
        return self._get_val_bool(_LU_THIN, default=None)

    def _is_thin(self, val):
        """val is boolean."""
        self.set_parm_value(_LU_THIN, u.sanitize_bool_for_api(val))

    @property
    def cloned_from_udid(self):
        return self._get_val_str(_LU_CLONED_FROM)

    def _cloned_from_udid(self, val):
        self.set_parm_value(_LU_CLONED_FROM, val)

    @property
    def in_use(self):
        return self._get_val_bool(_LU_IN_USE, default=None)


@ewrap.ElementWrapper.pvm_type('LogicalUnit', has_metadata=True,
                               child_order=_LU_EL_ORDER)
class LU(_LUBase, ewrap.ElementWrapper):
    """ElementWrapper representing a LogicalUnit DETAIL object.

    LogicalUnit exists as a DETAIL object e.g. within a SharedStoragePool
    (accessed via SSP.logical_units[n]) or a SCSI mapping (accessed via
    VIOS.scsi_mappings[n].backing_storage).
    """
    pass


@ewrap.EntryWrapper.pvm_type('LogicalUnit', child_order=_LU_EL_ORDER)
class LUEnt(_LUBase, ewrap.EntryWrapper):
    """EntryWrapper representing a LogicalUnit as a first-class REST object.

    LogicalUnit exists as a CHILD REST object under Tier.  This class provides
    the ability to perform e.g.

        LUEnt.get(adapter, parent=tier)
    """
    pass


@ewrap.EntryWrapper.pvm_type('Tier')
class Tier(ewrap.EntryWrapper):
    """A storage grouping within a SharedStoragePool."""

    @property
    def name(self):
        return self._get_val_str(_TIER_NAME)

    @property
    def udid(self):
        return self._get_val_str(_TIER_UDID)

    @property
    def is_default(self):
        return self._get_val_bool(_TIER_IS_DEFAULT)

    @property
    def capacity(self):
        return self._get_val_float(_TIER_CAPACITY)

    @property
    def ssp_uuid(self):
        """The UUID of this Tier's parent SharedStoragePool."""
        return u.get_req_path_uuid(self.get_href(_TIER_ASSOC_SSP,
                                                 one_result=True))


@ewrap.EntryWrapper.pvm_type('SharedStoragePool')
class SSP(ewrap.EntryWrapper):
    """A Shared Storage Pool containing PVs and LUs."""

    search_keys = dict(name='StoragePoolName')

    @classmethod
    def bld(cls, adapter, name, data_pv_list):
        """Create a fresh SSP EntryWrapper.

        :param adapter: A pypowervm.adapter.Adapter (for traits, etc.)
        :param name: String name for the SharedStoragePool.
        :param data_pv_list: Iterable of storage.PV instances
                             representing the data volumes for the
                             SharedStoragePool.
        """
        ssp = super(SSP, cls)._bld(adapter)
        # Assignment order matters.
        ssp.physical_volumes = data_pv_list
        ssp.name = name
        return ssp

    @property
    def name(self):
        return self._get_val_str(_SSP_NAME)

    @name.setter
    def name(self, newname):
        self.set_parm_value(_SSP_NAME, newname)

    @property
    def udid(self):
        return self._get_val_str(_SSP_UDID)

    @property
    def capacity(self):
        """Capacity in GB as a float."""
        return self._get_val_float(_SSP_CAPACITY)

    @property
    def free_space(self):
        """Free space in GB as a float."""
        return self._get_val_float(_SSP_FREE_SPACE)

    @property
    def over_commit_space(self):
        """Over commit space in GB as a float."""
        return self._get_val_float(_SSP_OCS)

    @property
    def total_lu_size(self):
        """Total LU size in GB as a float."""
        return self._get_val_float(_SSP_TOTAL_LU_SIZE)

    @property
    def logical_units(self):
        """WrapperElemList of LU wrappers."""
        return ewrap.WrapperElemList(self._find_or_seed(_SSP_LUS), LU)

    @logical_units.setter
    def logical_units(self, lus):
        self.replace_list(_SSP_LUS, lus)

    @property
    def physical_volumes(self):
        """WrapperElemList of PV wrappers."""
        return ewrap.WrapperElemList(self._find_or_seed(_SSP_PVS), PV)

    @physical_volumes.setter
    def physical_volumes(self, pvs):
        self.replace_list(_SSP_PVS, pvs)


@ewrap.Wrapper.base_pvm_type
class _VStorageAdapterMethods(ewrap.Wrapper):
    """Mixin to be used with _VStorageAdapter{Element|Entry}."""

    @property
    def side(self):
        """Will return either Server or Client.

        A Server indicates that this is a virtual adapter that resides on the
        Virtual I/O Server.

        A Client indicates that this is an adapter residing on a Client LPAR.
        """
        return self._get_val_str(_VADPT_TYPE)

    def _side(self, t):
        self.set_parm_value(_VADPT_TYPE, t)

    @property
    def is_varied_on(self):
        """True if the adapter is varied on."""
        return self._get_val_str(_VADPT_VARIED_ON)

    def _use_next_slot(self, use):
        """Use next available (not high) slot."""
        self.set_parm_value(_VADPT_NEXT_SLOT, u.sanitize_bool_for_api(use))

    @property
    def loc_code(self):
        """The device's location code."""
        return self._get_val_str(_VADPT_LOC_CODE)


# base_pvm_type by _VStorageAdapterMethods
@six.add_metaclass(abc.ABCMeta)
class _VStorageAdapterElement(ewrap.ElementWrapper, _VStorageAdapterMethods):
    """Parent class for the virtual storage adapters (FC or SCSI)."""
    has_metadata = True

    @classmethod
    def _bld_new(cls, adapter, side):
        """Build a {Client|Server}Adapter requesting a new virtual adapter.

        :param adapter: A pypowervm.adapter.Adapter (for traits, etc.)
        :param side: Either 'Client' or 'Server'.
        :returns: A fresh ClientAdapter or ServerAdapter wrapper with
                  UseNextAvailableSlotID=true
        """
        adp = super(_VStorageAdapterElement, cls)._bld(adapter)
        adp._use_next_slot(True)
        adp._side(side)
        return adp


# base_pvm_type by _VStorageAdapterMethods
@six.add_metaclass(abc.ABCMeta)
class _VStorageAdapterEntry(ewrap.EntryWrapper, _VStorageAdapterMethods):
    """Parent class for the virtual storage adapters (FC or SCSI)."""
    has_metadata = True

    @classmethod
    def _bld_new(cls, adapter, side):
        """Build a {Client|Server}Adapter requesting a new virtual adapter.

        :param adapter: A pypowervm.adapter.Adapter (for traits, etc.)
        :param side: Either 'Client' or 'Server'.
        :returns: A fresh ClientAdapter or ServerAdapter wrapper with
                  UseNextAvailableSlotID=true
        """
        adp = super(_VStorageAdapterEntry, cls)._bld(adapter)
        adp._side(side)
        adp._use_next_slot(True)
        return adp


@ewrap.Wrapper.base_pvm_type
class _VClientAdapterMethods(ewrap.Wrapper):
    """Mixin to be used with _VClientStorageAdapter{Element|Entry}."""
    @classmethod
    def bld(cls, adapter, slot_num=None):
        """Builds a new Client Adapter.

        If the slot number is None then we'll specify the
        'UseNextAvailableSlot' tag to REST and the REST layer will assign the
        slot. The slot number that it chose will be in the response.

        :param adapter: A pypowervm.adapter.Adapter
        :param slot_num: (Optional, Default: None) The client slot number to
                         be used.
        :returns: A new Client Adapter.
        """
        clad = super(_VClientAdapterMethods, cls)._bld_new(adapter, 'Client')
        if slot_num is not None:
            clad._lpar_slot_num(slot_num)
            clad._use_next_slot(False)
        return clad

    @property
    def lpar_id(self):
        """The short ID (not UUID) of the LPAR side of this adapter.

        Note that the LPAR ID is LocalPartitionID on the client side, and
        RemoteLogicalPartitionID on the server side.
        """
        return self._get_val_int(_VADPT_LOCAL_ID)

    @property
    def lpar_slot_num(self):
        """The (int) slot number that the adapter is in."""
        return self._get_val_int(_VADPT_SLOT_NUM)

    def _lpar_slot_num(self, slot_num):
        """Set the slot number that the adapter is in."""
        self.set_parm_value(_VADPT_SLOT_NUM, slot_num)


@six.add_metaclass(abc.ABCMeta)
@ewrap.ElementWrapper.pvm_type(CLIENT_ADPT, has_metadata=True,
                               child_order=_V_CLNT_ADPT_EL_ORDER)
class VClientStorageAdapterElement(_VClientAdapterMethods,
                                   _VStorageAdapterElement):
    """Parent class for Client Virtual Storage Adapter Elements."""
    pass


@six.add_metaclass(abc.ABCMeta)
@ewrap.ElementWrapper.pvm_type(SERVER_ADPT, has_metadata=True,
                               child_order=_V_SVR_ADPT_EL_ORDER)
class VServerStorageAdapterElement(_VStorageAdapterElement):
    """Parent class for Server Virtual Storage Adapters."""

    @classmethod
    def bld(cls, adapter):
        return super(VServerStorageAdapterElement, cls)._bld_new(adapter,
                                                                 'Server')

    @property
    def name(self):
        """The adapter's name on the Virtual I/O Server."""
        return self._get_val_str(_VADPT_NAME)

    @property
    def udid(self):
        """The device's Unique Device Identifier."""
        return self._get_val_str(_VADPT_UDID)


# pvm_type decorator by superclass (it is not unique)
class VSCSIClientAdapterElement(VClientStorageAdapterElement):
    """The Virtual SCSI Client Adapter within a VSCSI mapping.

    Paired with a VSCSIServerAdapterElement.
    """
    @property
    def vios_id(self):
        """The short ID (not UUID) of the VIOS side of this adapter.

        Note that the VIOS ID is RemoteLogicalPartitionID on the client side,
        and LocalPartitionID on the server side.
        """
        return self._get_val_int(_VSCSI_ADPT_REM_LPAR_ID)

    @property
    def vios_slot_num(self):
        """The (int) remote slot number of the paired adapter."""
        return self._get_val_int(_VSCSI_ADPT_REM_SLOT_NUM)


# pvm_type decorator by superclass (it is not unique)
class VSCSIServerAdapterElement(VServerStorageAdapterElement):
    """The Virtual SCSI Server Adapter within a VSCSI mapping.

    Paired with a VSCSIClientAdapterElement.
    """

    @property
    def backing_dev_name(self):
        """The backing device name that this virtual adapter is hooked into."""
        return self._get_val_str(_VSCSI_ADPT_BACK_DEV_NAME)

    @property
    def lpar_id(self):
        """The short ID (not UUID) of the LPAR side of this adapter.

        Note that the LPAR ID is LocalPartitionID on the client side, and
        RemoteLogicalPartitionID on the server side.
        """
        return self._get_val_int(_VSCSI_ADPT_REM_LPAR_ID)

    @property
    def vios_id(self):
        """The short ID (not UUID) of the VIOS side of this adapter.

        Note that the VIOS ID is RemoteLogicalPartitionID on the client side,
        and LocalPartitionID on the server side.
        """
        return self._get_val_int(_VADPT_LOCAL_ID)

    @property
    def lpar_slot_num(self):
        """The (int) slot number that the LPAR side of the adapter."""
        return self._get_val_int(_VSCSI_ADPT_REM_SLOT_NUM)

    @property
    def vios_slot_num(self):
        """The (int) slot number of the VIOS side of the adapter."""
        return self._get_val_int(_VADPT_SLOT_NUM)


@ewrap.Wrapper.base_pvm_type
class _VFCClientAdapterMethods(ewrap.Wrapper):
    """Mixin to be used with VFCClientAdapter(Element)."""
    def _wwpns(self, value):
        """Sets the WWPN string.

        :param value: The list of WWPNs.  Should only contain two.
        """
        if value is not None:
            self.set_parm_value(_VFC_CLNT_ADPT_WWPNS, " ".join(value).lower())

    @property
    def wwpns(self):
        """Returns a list that contains the WWPNs.  If no WWPNs, empty list."""
        val = self._get_val_str(_VFC_CLNT_ADPT_WWPNS)
        if val is None:
            return []
        else:
            return val.upper().split(' ')

    @property
    def vios_id(self):
        """The short ID (not UUID) of the VIOS side of this adapter."""
        return self._get_val_int(_VFC_ADPT_CONN_PARTITION_ID)

    @property
    def vios_slot_num(self):
        """The (int) remote slot number of the paired adapter."""
        return self._get_val_int(_VFC_ADPT_CONN_SLOT_NUM)


# pvm_type decorator by superclass (it is not unique)
class VFCClientAdapterElement(VClientStorageAdapterElement,
                              _VFCClientAdapterMethods):
    """The Virtual Fibre Channel Client Adapter within a VFC mapping.

    Paired with a VFCServerAdapterElement.
    """

    @classmethod
    def bld(cls, adapter, wwpns=None, slot_num=None):
        """Create a fresh Virtual Fibre Channel Client Adapter.

        :param adapter: A pypowervm.adapter.Adapter (for traits, etc.)
        :param wwpns: An optional set of two client WWPNs to set on the
                      adapter.
        :param slot_num: An optional integer to be set as the Virtual
                         slot number.
        """
        adpt = super(VFCClientAdapterElement, cls).bld(adapter,
                                                       slot_num=slot_num)

        if wwpns is not None:
            adpt._wwpns(wwpns)

        return adpt


@six.add_metaclass(abc.ABCMeta)
@ewrap.EntryWrapper.pvm_type(VFC_CLIENT_ADPT, has_metadata=True)
class VFCClientAdapter(_VStorageAdapterEntry, _VClientAdapterMethods,
                       _VFCClientAdapterMethods):
    """EntryWrapper for VirtualFibreChannelClientAdapter CHILD.

    Use this to wrap LogicalPartition/{uuid}/VirtualFibreChannelClientAdapter.
    """
    pass


# pvm_type decorator by superclass (it is not unique)
class VFCServerAdapterElement(VServerStorageAdapterElement):
    """The Virtual Fibre Channel Server Adapter within a VFC mapping.

    Paired with a VFCClientAdapterElement.
    """

    @property
    def map_port(self):
        """The physical FC port name that this virtual port is connect to."""
        return self._get_val_str(_VFC_SVR_ADPT_MAP_PORT)

    @property
    def lpar_id(self):
        """The short ID (not UUID) of the LPAR side of this adapter."""
        return self._get_val_int(_VFC_ADPT_CONN_PARTITION_ID)

    @property
    def vios_id(self):
        """The short ID (not UUID) of the VIOS side of this adapter."""
        return self._get_val_int(_VADPT_LOCAL_ID)

    @property
    def lpar_slot_num(self):
        """The (int) slot number that the LPAR side of the adapter."""
        return self._get_val_int(_VFC_ADPT_CONN_SLOT_NUM)

    @property
    def vios_slot_num(self):
        """The (int) slot number of the VIOS side of the adapter."""
        return self._get_val_int(_VADPT_SLOT_NUM)



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Objects that contain the per LPAR monitor data."""

import abc

import six


class LparMetric(object):
    """Represents a set of metrics for a given LPAR.

    This is a reduction and consolidation of the raw PCM statistics.
    """

    def __init__(self, uuid):
        """Creates a LPAR Metric.  Data will be set by invoker.

         - uuid - The LPAR's UUID.
         - memory - The LPAR's memory statistics.
         - processor - The LPAR's processor statistics.
         - network - The LparNetwork aggregation of network statistics.
         - storage - the LparStorage aggregation of storage statistics.

        If certain attributes are None, that means the statistics could not
        be pulled.

        :param uuid: The LPAR UUID
        """
        self.uuid = uuid
        self.memory = None
        self.processor = None
        self.network = None
        self.storage = None


@six.add_metaclass(abc.ABCMeta)
class PropertyWrapper(object):
    """Provides a thin wrapper around the raw metrics class.

    Sub class should have the _supported_metrics element defined.
    """

    def __init__(self, elem):
        self.elem = elem

    def __getattr__(self, attr):
        if attr not in self._supported_metrics:
            raise AttributeError()
        return getattr(self.elem, attr)


class LparMemory(object):
    """Represents the memory for a given LPAR.

    Requires the following as inputs:
      - PhypLparMemory raw metric
      - LparInfo.LparUtil raw metric. These metrics are got from IBM.Host
                           Resource Manager through RMC.

    The supported metrics are as follows:
      - logical_mem: The amount of memory on the LPAR.
      - backed_physical_mem: The amount of backing physical memory used by
                             the LPAR.
      - pct_real_mem_avbl: Percentage of available memory on VMs. It is
                           only available for newer RSCT packages.
                           This statistic does not count cached memory as
                           in use.
      - total_pg_count: Page count of swap space for this VM. Page
                        size is 4k.
      - free_pg_count: Page count of free swap space for this VM.
                       Page size is 4k.
      - active_pg_count: Page count of total active memory for this VM.
                         Page size is 4k.
      - real_mem_size_bytes: Total amount of memory assigned to this VM in
                             bytes.
      - pct_real_mem_free: Percentage of real page frames that are currently
                           available on the VMM (Virtual Memory Manager)
                           free list. VMM manages the allocation of real
                           memory page frames, resolves references to virtual
                           memory pages that are not currently in real memory
                           and manages the reading and writing of pages to
                           disk storage.
      - vm_pg_in_rate: Represents the rate (in pages per second) that the VMM
                           is reading both persistent and working pages from
                           disk storage. A -1 value indicates that system
                           could not determine this metric.
      - vm_pg_out_rate: Represents the rate (in pages per second) that the VMM
                           is writing both persistent and working pages to
                           disk storage. A -1 value indicates that system
                           could not determine this metric.
      - vm_pg_swap_in_rate: Represents the rate (in pages per second) that the
                           VMM is reading working pages from paging-space
                           disk storage. A -1 value indicates that system
                           could not determine this metric.
      - vm_pg_swap_out_rate: Represents the rate (in pages per second) that the
                           VMM is writing working pages to paging-space
                           disk storage. A -1 value indicates that system
                           could not determine this metric.
    """

    def __init__(self, lpar_mem_phyp, lpar_mem_pcm):
        self.logical_mem = lpar_mem_phyp.logical_mem
        self.backed_physical_mem = lpar_mem_phyp.backed_physical_mem
        # Its possible that for the lpar_sample, the memory metric was not
        # collected. If the metric is not available,
        # then assume 0 i.e. all memory is being utilized.
        if lpar_mem_pcm:
            self.pct_real_mem_avbl = lpar_mem_pcm.memory.pct_real_mem_avbl
            self.total_pg_count = lpar_mem_pcm.memory.total_pg_count
            self.free_pg_count = lpar_mem_pcm.memory.free_pg_count
            self.active_pg_count = lpar_mem_pcm.memory.active_pg_count
            self.real_mem_size_bytes = lpar_mem_pcm.memory.real_mem_size_bytes
            self.pct_real_mem_free = lpar_mem_pcm.memory.pct_real_mem_free
            self.vm_pg_in_rate = lpar_mem_pcm.memory.vm_pg_in_rate
            self.vm_pg_out_rate = lpar_mem_pcm.memory.vm_pg_out_rate
            self.vm_pg_swap_in_rate = lpar_mem_pcm.memory.vm_pg_swap_in_rate
            self.vm_pg_swap_out_rate = lpar_mem_pcm.memory.vm_pg_swap_out_rate
        else:
            self.pct_real_mem_free = 0
            self.vm_pg_in_rate = -1
            self.vm_pg_out_rate = -1
            self.vm_pg_swap_in_rate = -1
            self.vm_pg_swap_out_rate = -1


class LparProc(PropertyWrapper):
    """Represents the CPU statistics for a given LPAR.

    Requires the PhypLparProc raw metric as input.

    The supported metrics are as follows:
      - pool_id: The CPU pool for this LPAR.
      - mode: The CPU mode.  Typically cap or uncap.
      - virt_procs: The number of virtual processors assigned to the LPAR.
      - proc_units: The number of proc units assigned to the LPAR.

          Ex. if virt_procs is 4 and proc_units is .4, then each virtual
          processor has .1 CPUs.
      - weight: The CPU weight for uncapped processors.

          This defines how aggressive this CPU should be when using unused
          cycles from other LPARs (as compared to other VMs that may also
          request those unused cycles).
      - entitled_proc_cycles: The entitled number of processor cycles.
      - util_cap_proc_cycles: The number of used processor cycles from its
                              capped capacity.
      - util_uncap_proc_cycles: The number of utilized processor cycles pulled
                                from uncap spare.
      - idle_proc_cycles: The CPU cycles spent idling.
      - donated_proc_cycles: The number of CPU cycles donated to other VMs due
                             to no need.
      - time_wait_dispatch: Time spent waiting for CPU dispatch.
      - total_instructions: The total instructions executed.
      - total_inst_exec_time: The time for the instructions to execute.
    """

    _supported_metrics = ('pool_id', 'mode', 'virt_procs', 'proc_units',
                          'weight', 'entitled_proc_cycles',
                          'util_cap_proc_cycles', 'util_uncap_proc_cycles',
                          'idle_proc_cycles', 'donated_proc_cycles',
                          'time_wait_dispatch', 'total_instructions',
                          'total_inst_exec_time')


class LparStorage(object):
    """Represents the Storage statistics for a given LPAR.

    Requires the PhypLparStorage and list of ViosInfo raw metrics as input.

    Contains the various LPAR storage statistic elements.
     - virt_adapters - List of LparVirtStorageAdpt on the LPAR
     - vfc_adpts - List of LparVFCAdpt on the LPAR
    """

    def __init__(self, lpar_phyp_storage, vios_metrics):
        """Fills the VM storage metrics from the raw PHYP/VIOS metrics.

        :param lpar_phyp_storage: The raw Phyp Storage object.
        :param vios_metrics: The list of Virtual I/O Server raw metrics that
                             are paired to the sample from the lpar_phyp
                             metrics.
        """
        # Add the various adapters.
        self.virt_adpts = []
        for vadpt in lpar_phyp_storage.v_stor_adpts:
            vio_adpt = self._find_vio_vstor_adpt(vadpt, vios_metrics)
            if vio_adpt is not None:
                self.virt_adpts.append(LparVirtStorageAdpt(vio_adpt))

        self.vfc_adpts = []
        for phyp_vfc_adpt in lpar_phyp_storage.v_fc_adpts:
            vfc_adpt = self._find_vio_vfc_adpt(phyp_vfc_adpt, vios_metrics)
            if vfc_adpt is not None:
                self.vfc_adpts.append(LparVFCAdpt(vfc_adpt))

    @staticmethod
    def _find_vio_vstor_adpt(phyp_vadpt, vios_metrics):
        """Finds the appropriate VIOS virtual storage adapter.

        For a given PHYP virtual adapter, PHYP only has a little bit of
        information about it.  Which VIOS is hosting it, and the slot.

        The VIOS metrics actually contain the information for that
        device.

        This method will look through all the VIOS samples to find the
        matching ViosStorageVAdpt for the given PhypStorageVAdpt.  If one
        can not be found, None is returned.

        :param phyp_vadpt: The PhypStorageVAdpt raw metric.
        :param vios_metrics: The list of ViosInfos.
        :return: The corresponding ViosStorageVAdpt from the ViosInfos
                 if one can be found.  None otherwise.
        """
        for vios_ltm in vios_metrics:
            # We need to find the VIOS sample that matches this storage
            # element.  Loop until we find one (if one doesn't exist, then
            # this will just return None).
            if vios_ltm.sample.id != phyp_vadpt.vios_id:
                continue

            # If we reach here, we found the VIOS.  From that sample, see
            # if we have the appropriate storage.
            raw_stor = vios_ltm.sample.storage
            if raw_stor is None or raw_stor.virt_adpts is None:
                break

            # See if this virtual adapters has the right data.
            slot_str = "-C%d" % phyp_vadpt.vios_slot
            for vadpt in raw_stor.virt_adpts:
                # We have to match on the location code.  We can only match
                # on the tail end of the slot (we've already validated that
                # we have the right VIOS, so slot is sufficient).
                if vadpt.physical_location.endswith(slot_str):
                    return vadpt

            # If we reached this point, we found the right VIOS, but couldn't
            # find proper data.  Therefore we can just exit the loop.
            break

        return None

    @staticmethod
    def _find_vio_vfc_adpt(phyp_vfc_adpt, vios_metrics):
        """Finds the appropriate VIOS virtual FC adapter.

        For a given PHYP virtual FC adapter, PHYP only has a little bit of
        information about it.  Which VIOS is hosting it, and the WWPNs.

        The VIOS metrics actually contain the information for that
        device.

        This method will look through all the VIOS samples to find the
        matching ViosFCVirtAdpt for the given PhypVirtualFCAdpt.  If one
        can not be found, None is returned.

        :param phyp_vadpt: The PhypVirtualFCAdpt raw metric.
        :param vios_metrics: The list of ViosInfos.
        :return: The corresponding ViosFCVirtAdpt from the ViosInfos
                 if one can be found.  None otherwise.
        """
        for vios_ltm in vios_metrics:
            # We need to find the VIOS sample that matches this VFC
            # element.  Loop until we find one (if one doesn't exist, then
            # this will just return None).
            if vios_ltm.sample.id != phyp_vfc_adpt.vios_id:
                continue

            # If we reach here, we found the VIOS.  From that sample, see
            # if we have the appropriate storage.
            raw_stor = vios_ltm.sample.storage
            if raw_stor is None or raw_stor.fc_adpts is None:
                return None

            # Check the WWPNs.
            for pfc_adpt in raw_stor.fc_adpts:
                vfc_adpt = LparStorage._find_vfc(phyp_vfc_adpt, pfc_adpt)
                if vfc_adpt is not None:
                    return vfc_adpt

        return None

    @staticmethod
    def _find_vfc(phyp_vfc_adpt, vio_pfc_adpt):
        """Finds the matching VIOS vfc adpt for a given PHYP adapter

        :param phyp_vfc_adpt: The raw PhypVirtualFCAdpt object.
        :param vio_pfc_adpt: The raw ViosFCPhysAdpt.
        :return: The matching ViosFCVirtAdpt contained within the physical VIOS
                 adapter.  If one can't be found, None will be returned.
        """
        if vio_pfc_adpt.ports is None:
            return None
        for vfc_adpt in vio_pfc_adpt.ports:
            for wwpn in phyp_vfc_adpt.wwpn_pair:
                if wwpn == vfc_adpt.wwpn:
                    return vfc_adpt
        return None


@six.add_metaclass(abc.ABCMeta)
class LparStorageAdpt(PropertyWrapper):
    """Base class for storage adapters on a given LPAR.

    Requires the vios storage adapter raw metric as input.  Specific classes
    are defined by the subclasses.

    The supported metrics are as follows:
      - name: The identifier of the adapter.  Ex: vhost2.
      - physical_location: The physical location code of the adapter.
      - num_reads: The number of read operations done against the adapter.
      - num_writes: The number of write operations done against the adapter.
      - read_bytes: The number of bytes read from the adapter.
      - write_bytes: The number of bytes written to the adapter.
      - type: The type of the adapter.
    """

    _supported_metrics = ('name', 'physical_location', 'num_reads', 'type',
                          'num_writes', 'read_bytes', 'write_bytes')


class LparVFCAdpt(LparStorageAdpt):
    """A Virtual Fibre Channel Adapter attached to the LPAR.

    Requires the ViosFCVirtAdpt raw metric as input.

    The supported metrics are as follows:
      - name: The identifier of the adapter.  Ex: vhost2.
      - physical_location: The physical location code of the adapter.
      - num_reads: The number of read operations done against the adapter.
      - num_writes: The number of write operations done against the adapter.
      - read_bytes: The number of bytes read from the adapter.
      - write_bytes: The number of bytes written to the adapter.
      - type: The type of the adapter.  Will be set to VFC.
    """

    @property
    def type(self):
        """Overrides the type property as the raw metric.

        The VFC Adapter does not natively have a type in the raw metric.  This
        property overrides and circumvents the standard property lookup
        mechanism.
        """
        return "VFC"


class LparPhysAdpt(LparStorageAdpt):
    """A physical adapter (ex SAS drive) on the LPAR.

    Requires the ViosStoragePAdpt raw metric as input.

    The supported metrics are as follows:
      - name: The identifier of the adapter.  Ex: vhost2.
      - physical_location: The physical location code of the adapter.
      - num_reads: The number of read operations done against the adapter.
      - num_writes: The number of write operations done against the adapter.
      - read_bytes: The number of bytes read from the adapter.
      - write_bytes: The number of bytes written to the adapter.
      - type: The type of the adapter.
    """
    pass


class LparVirtStorageAdpt(LparStorageAdpt):
    """A Virutal Storage Adapter (ex. vscsi) attached to the LPAR.

    Requires the ViosStorageVAdpt raw metric as input.

    The supported metrics are as follows:
      - name: The identifier of the adapter.  Ex: vhost2.
      - physical_location: The physical location code of the adapter.
      - num_reads: The number of read operations done against the adapter.
      - num_writes: The number of write operations done against the adapter.
      - read_bytes: The number of bytes read from the adapter.
      - write_bytes: The number of bytes written to the adapter.
      - type: The type of the adapter.
    """
    pass


class LparNetwork(object):
    """Represents the Network statistics for a given LPAR.

    Requires the PhypNetwork raw metric as input.

    Aggregates the various types of network statistics for a given LPAR.
     - cnas - List of the Client Network Adapter stats.
    """

    def __init__(self, lpar_sample_net):
        """Creates the Network Statistics aggregation element.

        Puts the network information into the lpar_metric.network variable.

        :param lpar_sample_net: The PHYP raw data sample.
        """
        # Fill in the Client Network Adapter data sources
        self.cnas = ([] if lpar_sample_net.veas is None else
                     [LparCNA(x) for x in lpar_sample_net.veas])

        # TODO(thorst) Additional network metrics.  Ex. SR-IOV ports


class LparCNA(PropertyWrapper):
    """Represents a Client Network Adapter on a LPAR.

    Requires the PhypVEA raw metric as input.

    The supported metrics are as follows:
      - vlan_id: The PVID of the Client Network Adapter.
      - vswitch_id: The virtual switch ID (not UUID).
      - physical_location: The physical location for the Client Network
                           Adapter.
      - received_packets: The count of packets received to the Client Network
                          Adapter.
      - sent_packets: The count of packets sent by the Client Network Adapter.
      - dropped_packets: The count of the packets dropped by the Client Network
                         Adapter.
      - sent_bytes: The count of the bytes sent by the Client Network Adapter.
      - received_bytes: The count of the bytes received by the Client Network
                        Adapter.
    """

    _supported_metrics = ('vlan_id', 'vswitch_id', 'physical_location',
                          'received_packets', 'sent_packets',
                          'dropped_packets', 'sent_bytes', 'received_bytes')



# Copyright 2017 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Helper classes for PowerOn/PowerOff options (additional Job parameters)."""

import abc
import six

import pypowervm.exceptions as exc
import pypowervm.wrappers.base_partition as bp
from pypowervm.wrappers import job
import pypowervm.wrappers.logical_partition as lpar


IPLSrc = lpar.IPLSrc


class BootMode(object):
    """Valid values for the 'bootmode' parameter in power_on.

    Not to be confused with pypowervm.wrappers.base_partition.BootMode.

    Example usage:
        power_on(..., add_parms={BootMode.KEY: BootMode.SMS, ...})
    """
    KEY = 'bootmode'
    NORM = 'norm'
    SMS = 'sms'
    DD = 'dd'
    DS = 'ds'
    OF = 'of'
    ALL_VALUES = (NORM, SMS, DD, DS, OF)


class KeylockPos(object):
    """Valid values for the 'keylock' parameter in power_on.

    Not to be confused with pypowervm.wrappers.base_partition.KeylockPos.

    Example usage:
        power_on(..., add_parms={KeylockPos.KEY: KeylockPos.MANUAL, ...})
    """
    KEY = 'keylock'
    MANUAL = 'manual'
    NORMAL = 'norm'
    UNKNOWN = 'unknown'
    ALL_VALUES = (MANUAL, NORMAL, UNKNOWN)


class RemoveOptical(object):
    """Valid values for the 'remove_optical_*' parameters in power_on.

    This is primarily used to remove the config drive after install. KEY_NAME
    is required and maps to a VirtualOpticalMedia name to remove. KEY_TIME is
    optional and maps to the time, in minutes, to wait before deleting the
    media.

    Example usage:
        power_on(..., add_parms={RemoveOptical.KEY_TIME: <Integer>,
                                 RemoveOptical.KEY_NAME: <String>}, ...)
    """
    KEY_NAME = 'remove_optical_name'
    KEY_TIME = 'remove_optical_time'

    @classmethod
    def bld_map(cls, name, time=0):
        return {cls.KEY_NAME: name, cls.KEY_TIME: time}


class IBMiOperationType(object):
    """Valid values for the IBMi operation type in power_on."""
    KEY = 'OperationType'
    ACTIVATE = 'activate'
    NETBOOT = 'netboot'
    CHANGE_KEYLOCK = 'changeKeylock'
    ALL_VALUES = (ACTIVATE, NETBOOT, CHANGE_KEYLOCK)


class PowerOffOperation(object):
    """Valid values for the operation in power_off."""
    KEY = 'operation'
    VSP = 'shutdown'
    OS = 'osshutdown'
    DUMPRESTART = 'dumprestart'
    ALL_VALUES = (VSP, OS, DUMPRESTART)


class Force(object):
    """Enumeration indicating the strategy for forcing power-off."""
    # The force-immediate option is included on the first pass.
    TRUE = True
    # The force-immediate option is not included on the first pass; but if the
    # power-off fails, it is retried with the force-immediate option included.
    # This value is False for backward compatibility.
    ON_FAILURE = False
    # The force-immediate option is not included.  If the power-off fails, it
    # is not retried.
    NO_RETRY = 'no retry'


@six.add_metaclass(abc.ABCMeta)
class _PowerOpts(object):
    def __init__(self, legacy_add_parms=None):
        """Initialize a PowerOpts instance.

        :param legacy_add_parms: For legacy use only, initialize the internal
                                 parameter map from the specified dictionary of
                                 Job parameter name/value pairs.
        """
        self._parm_map = {}
        if legacy_add_parms is not None:
            self._parm_map.update(legacy_add_parms)

    def __str__(self):
        """String representation of this instance, for log/test purposes."""
        parms = ', '.join(["%s=%s" % (key, self._parm_map[key])
                           for key in sorted(self._parm_map)])
        return "%s(%s)" % (self.JOB_SUFFIX, parms)

    def _process_enum(self, enum, value):
        if value not in enum.ALL_VALUES:
            raise exc.InvalidEnumValue(enum=enum.__name__, value=value,
                                       valid_values=str(enum.ALL_VALUES))
        self._parm_map[enum.KEY] = value
        return self

    def _process_bool(self, key, value):
        """Process a boolean option.

        All boolean options are false by default.  Thus, if value is 'true'/i
        or True, the key is added with the value 'true'; otherwise it is
        *removed* from the _PowerOpt.

        :param key: The JobParameterName.
        :param value: A bool (True/False) or string ('true', 'false',
                      case-insensitive).  Default: True.
        """
        if key in self._parm_map:
            del self._parm_map[key]
        if str(value).lower() == 'true':
            self._parm_map[key] = 'true'
        return self

    def bld_jparms(self):
        return [job.Job.create_job_parameter(key, str(val)) for key, val in
                six.iteritems(self._parm_map)]


class PowerOnOpts(_PowerOpts):
    """Job parameters for pypowervm.tasks.power.power_on/PowerOp.start."""

    JOB_SUFFIX = 'PowerOn'

    def bootmode(self, value):
        """Set the boot mode.

        :param value: One of the BootMode enum values.
        :return self: For chaining.
        """
        return self._process_enum(BootMode, value)

    def keylock_pos(self, value):
        """Set the Keylock Position.

        :param value: One of the KeylockPos enum values.
        :return self: For chaining.
        """
        return self._process_enum(KeylockPos, value)

    def bootstring(self, value):
        """Set the boot string.

        :param value: The boot string to use.
        :return self: For chaining.
        """
        self._parm_map['bootstring'] = value
        return self

    def force(self, value=True):
        """Add the force option.

        :param value: A bool (True/False) or string ('true', 'false',
                      case-insensitive).  Default: True.
        :return self: For chaining.
        """
        return self._process_bool('force', value)

    def remove_optical(self, name, time=0):
        """Add options to remove an optical drive after boot.

        :param name: The name of a VirtualOpticalMedia name to remove.
        :param time: The time, in minutes, to wait before deleting the media.
        :return self: For chaining.
        """
        self._parm_map.update(RemoveOptical.bld_map(name, time=time))
        return self

    def ibmi_ipl_source(self, value):
        """Set the IBMi IPL Source.

        :param value: One of the IPLSrc enum values.
        :return self: For chaining.
        """
        return self._process_enum(IPLSrc, value)

    def ibmi_op_type(self, value):
        """Set the IBMi Operation Type.

        :param value: One of the IBMiOperationType enum values.
        :return self: For chaining.
        """
        return self._process_enum(IBMiOperationType, value)

    def ibmi_netboot_params(self, ipaddr, serverip, gateway, serverdir,
                            subnet=None, connspeed=None, duplex=None, mtu=None,
                            vlanid=None):
        """Set parameters for IBMi netboot.

        Use with  ibmi_op_type(IBMiOperationType.NETBOOT).

        :param ipaddr: IP (v4 or v6) address of the client VM.
        :param serverip: IP (v4 or v6) address of the netboot server.
        :param gateway: IP (v4 or v6) address of the gateway.
        :param serverdir: Location of the netboot image on the server.
        :param subnet: Subnet mask.  IPv4 only.
        :param connspeed: Connection speed.
        :param duplex: Duplex mode.
        :param mtu: Maximum Transmission Unit.
        :param vlanid: VLAN ID.
        :return self: For chaining.
        """
        self._parm_map['IPAddress'] = ipaddr
        self._parm_map['ServerIPAddress'] = serverip
        self._parm_map['Gateway'] = gateway
        self._parm_map['IBMiImageServerDirectory'] = serverdir
        # Optional args
        for key, val in (('SubnetMask', subnet),
                         ('ConnectionSpeed', connspeed),
                         ('DuplexMode', duplex),
                         ('VLANID', vlanid),
                         ('MaximumTransmissionUnit', mtu)):
            if val is not None:
                # connspeed/vlanid/mtu may arrive as ints
                self._parm_map[key] = str(val)
        return self


class PowerOffOpts(_PowerOpts):
    """Job parameters for pypowervm.tasks.power.power_off/PowerOp.stop.

    Use *one* of os_normal, os_immediate, vsp_normal, vsp_hard, or soft_detect.

    Optionally specify restart.
    """
    JOB_SUFFIX = 'PowerOff'

    def immediate(self, value=True):
        """Whether to include immediate=true.

        This corresponds to "hard" for VSP, "immediate" for OS.

        This should only be used with operation(DUMPRESTART).  Otherwise, one
        of the os_normal, os_immediate, vsp_normal, vsp_hard, or soft_detect
        methods should be used.

        :param value: A bool (True/False) or string ('true', 'false',
                      case-insensitive).
        :return self: For chaining.
        """
        return self._process_bool('immediate', value)

    @property
    def is_immediate(self):
        return self._parm_map.get('immediate') == 'true'

    def operation(self, value):
        """The PowerOff operation to perform.

        This should only be used for DUMPRESTART.  Otherwise, one of the
        os_normal, os_immediate, vsp_normal, vsp_hard, or soft_detect methods
        should be used.

        :param value: One of the PowerOffOperation enum values.
        :return self: For chaining.
        """
        return self._process_enum(PowerOffOperation, value)

    @staticmethod
    def can_os_shutdown(part):
        """Can the specified partition perform an OS shutdown?

        :param part: LPAR/VIOS wrapper indicating the partition to inspect.
        :return: True if the specified partition is capable of OS shutdown;
                 False otherwise.
        """
        # OS shutdown is always available on IBMi partitions.
        # OS shutdown is available if RMC is up.
        return (part.env == bp.LPARType.OS400) or (part.rmc_state ==
                                                   bp.RMCState.ACTIVE)

    @property
    def is_os(self):
        return self._parm_map.get('operation') == PowerOffOperation.OS

    def restart(self, value=True):
        """Whether to restart the partition after power-off.

        :param value: A bool (True/False) or string ('true', 'false',
                      case-insensitive).  Default: True.
        :return self: For chaining.
        """
        return self._process_bool('restart', value)

    @property
    def is_restart(self):
        return self._parm_map.get('restart') == 'true'

    def os_normal(self):
        """Set up normal OS shutdown.

        Sends the 'shutdown' command to the operating system.

        :return self: For chaining.
        """
        return self.operation(PowerOffOperation.OS).immediate(value=False)

    def os_immediate(self):
        """Set up immediate OS shutdown.

        Sends the 'shutdown -t now' command to the operating system.

        :return self: For chaining.
        """
        return self.operation(PowerOffOperation.OS).immediate()

    def vsp_normal(self):
        """Set up normal VSP shutdown.

        The Virtual Service Processor sends the equivalent of an EPOW event to
        the operating system.  The result is OS-dependent.

        :return self: For chaining.
        """
        return self.operation(PowerOffOperation.VSP).immediate(value=False)

    def vsp_hard(self):
        """Set up hard VSP shutdown.

        Akin to pulling the plug from the partition.  Processors are stopped
        immediately, and any pending I/O is lost.  May result in data
        corruption.

        :return self: For chaining.
        """
        return self.operation(PowerOffOperation.VSP).immediate()

    def soft_detect(self, part, immed_if_os=None):
        """Determine the appropriate soft shutdown operation for a partition.

        For IBMi partitions, this will always set up an OS shutdown.

        For non-IBMi partitions with active RMC, this will set up an OS
        shutdown.

        For non-IBMi partitions without RMC, this will set up a *normal* VSP
        shutdown.

        :param part: LPAR or VIOS wrapper indicating the partition being shut
                     down.
        :param immed_if_os: If this method determines that an OS shutdown is to
                            be performed, this parameter indicates whether that
                            shutdown should be immediate (True) or not (False).
                            The default is False for IBMi partitions, and True
                            for non-IBMi partitions.  This parameter is ignored
                            if a VSP shutdown is detected.
        :return self: For chaining.
        """
        if self.can_os_shutdown(part):
            self.operation(PowerOffOperation.OS)
            # Specific 'immediate' behavior requested for OS shutdown?
            if immed_if_os is not None:
                self.immediate(value=immed_if_os)
            else:
                # Default is normal for IBMi, immediate for non-IBMi.
                self.immediate(value=part.env != bp.LPARType.OS400)
        else:
            # OS shutdown not available; perform *normal* VSP shutdown.
            self.vsp_normal()
        return self



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This script attempts to create a real cluster on a real system using the
# cluster_ssp module.  Execute from project root directory as:
# PYTHONPATH=. python pypowervm/tests/tasks/create_cluster.py

import six

import pypowervm.adapter as adp
import pypowervm.exceptions as ex
import pypowervm.tasks.cluster_ssp as cs
import pypowervm.wrappers.cluster as clust
import pypowervm.wrappers.storage as stor

# >>>Replace the following with real values>>>
HOST = '9.1.2.3'
USER = 'hscroot'
PASS = 'abc123'

NODE_HOSTNAME = 'vios1.example.com'
NODE_MTMS = '8247-22L*1234D0A'
NODE_LPARID = 2
NODE_URI = ('https://9.1.2.3:12443/rest/api/uom/VirtualIOServer/'
            '58C9EB1D-7213-4956-A011-77D43CC4ACCC')
REPOS_UDID = '01M0lCTTIxNDUxMjQ2MDA1MDc2ODAyODI4NjFEODgwMDAwMDAwMDAwMDAwMg=='
REPOS_NAME = 'hdisk2'
DATA1_UDID = '01M0lCTTIxNDUxMjQ2MDA1MDc2ODAyODI4NjFEODgwMDAwMDAwMDAwMDAwMw=='
DATA1_NAME = 'hdisk3'
DATA2_UDID = '01M0lCTTIxNDUxMjQ2MDA1MDc2ODAyODI4NjFEODgwMDAwMDAwMDAwMDAwNA=='
DATA2_NAME = 'hdisk4'
DATA3_UDID = '01M0lCTTIxNDUxMjQ2MDA1MDc2ODAyODI4NjFEODgwMDAwMDAwMDAwMDAwNQ=='
DATA3_NAME = 'hdisk5'
# <<<Replace the foregoing with real values<<<

sess = adp.Session(HOST, USER, PASS, certpath=None)
adap = adp.Adapter(sess)

# Option 1: MTMS, LPAR_ID, Hostname
node1 = clust.Node.bld(adap, hostname=NODE_HOSTNAME, mtms=NODE_MTMS,
                       lpar_id=NODE_LPARID)

# Option 2: URI
node2 = clust.Node.bld(adap, vios_uri=NODE_URI)

repos = stor.PV.bld(adap, udid=REPOS_UDID, name=REPOS_NAME)

data_pvs = [
    stor.PV.bld(adap, udid=udid, name=name) for udid, name in (
        (DATA1_UDID, DATA1_NAME),
        (DATA2_UDID, DATA2_NAME),
        (DATA3_UDID, DATA3_NAME))]
try:
    cs.crt_cluster_ssp('clust1', 'ssp1', repos, node1, data_pvs)
except ex.JobRequestFailed as e:
    print(six.text_type(e))

adap = None



# Copyright 2014, 2017 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tasks to start, stop, and reboot partitions."""

from oslo_config import cfg
from oslo_log import log as logging
import six

import pypowervm.const as c
import pypowervm.exceptions as pexc
from pypowervm.i18n import _
import pypowervm.log as lgc
import pypowervm.tasks.power_opts as popts
import pypowervm.wrappers.base_partition as bp
from pypowervm.wrappers import job

LOG = logging.getLogger(__name__)

CONF = cfg.CONF

# Error codes indicate osshutdown is not supported
_OSSHUTDOWN_RMC_ERRS = ['HSCL0DB4', 'PVME01050905', 'PVME01050402']
# Error codes indicate partition is already powered off
_ALREADY_POWERED_OFF_ERRS = ['HSCL1558', 'PVME04000005', 'PVME01050901']
# Error codes indicate partition is already powered on
_ALREADY_POWERED_ON_ERRS = ['HSCL3681', 'PVME01042026']

BootMode = popts.BootMode
KeylockPos = popts.KeylockPos
RemoveOptical = popts.RemoveOptical
Force = popts.Force


class PowerOp(object):
    """Provides granular control over a partition PowerOn/Off Job.

    Use the start or stop @classmethod to invoke the appropriate Job.  Jobs
    invoked through these methods are never retried.  If they fail or time out,
    they raise relevant exceptions - see the methods' docstrings for details.
    """
    @classmethod
    def start(cls, part, opts=None, timeout=CONF.pypowervm_job_request_timeout,
              synchronous=True):
        """Power on a partition.

        :param part: Partition (LPAR or VIOS) wrapper indicating the partition
                     to power on.
        :param opts: An instance of power_opts.PowerOnOpts indicating
                     additional options to specify to the PowerOn operation.
                     By default, no additional options are used.
        :param timeout: value in seconds for specifying how long to wait for
                        the Job to complete.
        :param synchronous: If True, this method will not return until the Job
                            completes (whether success or failure) or times
                            out.  If False, this method will return as soon as
                            the Job has started on the server (that is,
                            achieved any state beyond NOT_ACTIVE).  Note that
                            timeout is still possible in this case.
        :raise VMPowerOnTimeout: If the Job timed out.
        :raise VMPowerOnFailure: If the Job failed for some reason other than
                                 that the partition was already powered on.
        """
        try:
            cls._run(part, opts or popts.PowerOnOpts(), timeout,
                     synchronous=synchronous)
        except pexc.JobRequestTimedOut as error:
            LOG.exception(error)
            raise pexc.VMPowerOnTimeout(lpar_nm=part.name, timeout=timeout)
        except pexc.JobRequestFailed as error:
            emsg = six.text_type(error)
            # If already powered on, don't send exception
            if (any(err_prefix in emsg
                    for err_prefix in _ALREADY_POWERED_ON_ERRS)):
                LOG.warning(_("Partition %s already powered on."), part.name)
                return
            LOG.exception(error)
            raise pexc.VMPowerOnFailure(lpar_nm=part.name, reason=emsg)

    @classmethod
    def stop(cls, part, opts=None, timeout=CONF.pypowervm_job_request_timeout,
             synchronous=True):
        """Power off a partition.

        :param part: LPAR/VIOS wrapper indicating the partition to power off.
        :param opts: An instance of power_opts.PowerOffOpts indicating the type
                     of shutdown to perform, and any additional options.  If
                     not specified, PowerOffOpts.soft_detect is used, with no
                     restart.
        :param timeout: value in seconds for specifying how long to wait for
                        the Job to complete.
        :param synchronous: If True, this method will not return until the Job
                            completes (whether success or failure) or times
                            out.  If False, this method will return as soon as
                            the Job has started on the server (that is,
                            achieved any state beyond NOT_ACTIVE).  Note that
                            timeout is still possible in this case.
        :raise VMPowerOffTimeout: If the Job timed out.
        :raise VMPowerOffFailure: If the Job failed for some reason other than
                                  that the partition was already powered off,
                                  and restart was not requested.
        :return: A PowerOp instance which can be invoked via the run method.
        :raise OSShutdownNoRMC: OP_PWROFF_OS was requested on a non-IBMi
                                partition with no RMC connection.
        """
        if opts is None:
            opts = popts.PowerOffOpts().soft_detect(part)

        if opts.is_os and not opts.can_os_shutdown(part):
            raise pexc.OSShutdownNoRMC(lpar_nm=part.name)

        try:
            cls._run(part, opts, timeout, synchronous=synchronous)
        except pexc.JobRequestTimedOut as error:
            LOG.exception(error)
            raise pexc.VMPowerOffTimeout(lpar_nm=part.name, timeout=timeout)
        except pexc.JobRequestFailed as error:
            emsg = six.text_type(error)
            # If already powered off and not a reboot, don't send exception
            if (any(err_prefix in emsg
                    for err_prefix in _ALREADY_POWERED_OFF_ERRS) and
                    not opts.is_restart):
                LOG.warning(_("Partition %s already powered off."), part.name)
                return
            LOG.exception(error)
            raise pexc.VMPowerOffFailure(lpar_nm=part.name, reason=emsg)

    @classmethod
    def _run(cls, part, opts, timeout, synchronous=True):
        """Fetch, fill out, and run a Power* Job for this PowerOp.

        Do not invoke this method directly; it is used by the start and stop
        class methods.

        :param part: LPAR/VIOS wrapper of the partition to power on/off.
        :param opts: Instance of power_opts.PowerOnOpts or PowerOffOpts
                     indicating the type of operation to perform, and any
                     additional options.
        :param timeout: value in seconds for specifying how long to wait for
                        the Job to complete.
        :param synchronous: If True, this method will not return until the Job
                            completes (whether success or failure) or times
                            out.  If False, this method will return as soon as
                            the Job has started on the server (that is,
                            achieved any state beyond NOT_ACTIVE).  Note that
                            timeout is still possible in this case.
        :raise VMPowerOffTimeout: If the Job timed out.
        :raise VMPowerOffFailure: If the Job failed.
        """
        # Fetch the Job template wrapper
        jwrap = job.Job.wrap(part.adapter.read(
            part.schema_type, part.uuid, suffix_type=c.SUFFIX_TYPE_DO,
            suffix_parm=opts.JOB_SUFFIX))

        LOG.debug("Executing power operation for partition %(lpar_nm)s with "
                  "timeout=%(timeout)d and synchronous=%(synchronous)s: "
                  "%(opts)s",
                  dict(lpar_nm=part.name, timeout=timeout,
                       synchronous=synchronous, opts=str(opts)))
        # Run the Job, letting exceptions raise up.
        jwrap.run_job(part.uuid, job_parms=opts.bld_jparms(), timeout=timeout,
                      synchronous=synchronous)


def _legacy_power_opts(klass, add_parms):
    """Detect (and warn) if add_parms is a legacy dict vs. a Power*Opts.

    Usage:
        opts, legacy = _legacy_power_opts(PowerOnOpts, add_parms)
        if legacy:
            # Do other stuff based on legacy behavior
        else:
            # Do other stuff based on new behavior

    :param klass: The class we expect, either PowerOnOpts or PowerOffOpts.
    :param add_parms: The add_parms argument to check.
    :return: An instance of klass, which is either add_parms, constructed by
             passing it to the klass __init__'s legacy_add_parms.
    :return: True if add_parms was a legacy dict; False otherwise.
    """
    if isinstance(add_parms, klass):
        return add_parms, False
    else:
        if add_parms is not None:
            import warnings
            warnings.warn(_("Specifying add_parms as a dict is deprecated. "
                            "Please specify a %s instance instead.") %
                          klass.__name__, DeprecationWarning)
        return klass(legacy_add_parms=add_parms), True


@lgc.logcall_args
def power_on(part, host_uuid, add_parms=None, synchronous=True):
    """Will Power On a Logical Partition or Virtual I/O Server.

    :param part: The LPAR/VIOS wrapper of the partition to power on.
    :param host_uuid: Not used.  Retained for backward compatibility.
    :param add_parms: A power_opts.PowerOnOpts instance; or (deprecated) a dict
                      of parameters to pass directly to the job template.  If
                      unspecified, a default PowerOnOpts instance is used, with
                      no additional parameters.
    :param synchronous: If True (the default), this method will not return
                        until the PowerOn Job completes (whether success or
                        failure) or times out.  If False, this method will
                        return as soon as the Job has started on the server
                        (that is, achieved any state beyond NOT_ACTIVE).  Note
                        that timeout is still possible in this case.
    :raise VMPowerOnFailure: If the operation failed.
    :raise VMPowerOnTimeout: If the operation timed out.
    """
    PowerOp.start(
        part, opts=_legacy_power_opts(popts.PowerOnOpts, add_parms)[0],
        synchronous=synchronous)


def _pwroff_soft_ibmi_flow(part, opts, timeout):
    """Normal (non-hard) power-off retry flow for IBMi partitions.

          ==================
          opts.is_immediate? <--START
          ==================
           |             |
           NO           YES
           V             V
    =========          ============          ==========          ============
    OS normal -FAIL*-> OS immediate -FAIL*-> VSP normal -FAIL*-> return False
    =========          ============          ==========          ============
        |_________________ | ___________________|
                          |||
                        SUCCESS
                           V         *VMPowerOffTimeout OR
                      ===========     VMPowerOffFailure
                      return True
                      ===========

    :param part timeout: See power_off.
    :param opts: A PowerOffOpts instance.  The operation and immediate params
                 are overwritten at each stage of this method.  Any other
                 options (such as restart) remain unaffected.
    :return: True if the power-off succeeded; False otherwise.  The caller
             (power_off) should perform VSP hard shutdown if False is returned.
    :raise VMPowerOffTimeout: If the last power-off attempt timed out.
    :raise VMPowerOffFailure: If the last power-off attempt failed.
    """
    # If immediate was already specified, skip OS-normal.
    if not opts.is_immediate:
        # ==> OS normal
        try:
            PowerOp.stop(part, opts=opts.os_normal(), timeout=timeout)
            return True
        except pexc.VMPowerOffFailure:
            LOG.warning(_("IBMi OS normal shutdown failed.  Trying OS "
                          "immediate shutdown.  Partition: %s"), part.name)
            # Fall through to OS immediate, with default timeout
            timeout = CONF.pypowervm_job_request_timeout

    # ==> OS immediate
    try:
        PowerOp.stop(part, opts=opts.os_immediate(), timeout=timeout)
        return True
    except pexc.VMPowerOffFailure:
        LOG.warning(_("IBMi OS immediate shutdown failed.  Trying VSP normal "
                      "shutdown.  Partition: %s"), part.name)
        # Fall through to VSP normal

    # ==> VSP normal
    try:
        PowerOp.stop(part, opts=opts.vsp_normal(), timeout=timeout)
        return True
    except pexc.VMPowerOffFailure:
        LOG.warning(_("IBMi VSP normal shutdown failed.  Trying VSP hard "
                      "shutdown.  Partition: %s"), part.name)

    return False


def _pwroff_soft_standard_flow(part, opts, timeout):
    """Normal (non-hard) power-off retry flow for non-IBMi partitions.

    START
    |     +---VMPowerOffTimeout-------------------------------------+
    V     |                                                         V
    ========  VMPowerOffFailure  ==========  VMPowerOffFailure  ============
    OS immed ----   or      ---> VSP normal ----   or      ---> return False
    ========   OSShutdownNoRMC   ==========  VMPowerOffTimeout  ============
          | _________________________/
          |/
       SUCCESS
          V
     ===========
     return True
     ===========

    :param part timeout: See power_off.
    :param opts: A PowerOffOpts instance.  The operation and immediate params
                 are overwritten at each stage of this method.  Any other
                 options (such as restart) remain unaffected.
    :return: True if the power-off succeeded; False otherwise.  The caller
             should perform VSP hard shutdown if False is returned.
    :raise VMPowerOffTimeout: If the last power-off attempt timed out.
    :raise VMPowerOffFailure: If the last power-off attempt failed.
    """
    # For backward compatibility, OS shutdown is always immediate.  We don't
    # let PowerOn decide whether to use OS or VSP; instead we trap
    # OSShutdownNoRMC (which is very quick) so we can keep this progression
    # linear.

    # ==> OS immediate
    try:
        PowerOp.stop(part, opts=opts.os_immediate(), timeout=timeout)
        return True
    except pexc.VMPowerOffTimeout:
        LOG.warning(_("Non-IBMi OS immediate shutdown timed out.  Trying VSP "
                      "hard shutdown.  Partition: %s"), part.name)
        return False
    except pexc.VMPowerOffFailure:
        LOG.warning(_("Non-IBMi OS immediate shutdown failed.  Trying VSP "
                      "normal shutdown.  Partition: %s"), part.name)
        # Fall through to VSP normal, but with default timeout
        timeout = CONF.pypowervm_job_request_timeout
    except pexc.OSShutdownNoRMC as error:
        LOG.warning(error.args[0])
        # Fall through to VSP normal

    # ==> VSP normal
    try:
        PowerOp.stop(part, opts.vsp_normal(), timeout=timeout)
        return True
    except pexc.VMPowerOffFailure:
        LOG.warning(_("Non-IBMi VSP normal shutdown failed.  Trying VSP hard "
                      "shutdown.  Partition: %s"), part.name)

    return False


def _power_off_single(part, opts, force_immediate, timeout):
    """No-retry single power-off operation.

    :param part force_immediate timeout: See power_off.
                                    force_immediate is either TRUE or NO_RETRY.
    :param opts: A PowerOffOpts instance.  The operation and immediate params
                 are overwritten by this method.  Any other options (such as
                 restart) remain unaffected.
    :raise VMPowerOffFailure: If the operation failed.
    :raise VMPowerOffTimeout: If the operation timed out.
    """
    # If force_immediate=TRUE, always VSP hard shutdown.
    if force_immediate == Force.TRUE:
        PowerOp.stop(part, opts=opts.vsp_hard(), timeout=timeout)
    # If no retries, just do the single "soft" power-off requested
    elif force_immediate == Force.NO_RETRY:
        # opts is already set up for soft_detect
        PowerOp.stop(part, opts=opts, timeout=timeout)

    return


@lgc.logcall_args
def power_off(part, host_uuid, force_immediate=Force.ON_FAILURE, restart=False,
              timeout=CONF.pypowervm_job_request_timeout, add_parms=None):
    """Will Power Off a Logical Partition or Virtual I/O Server.

    Depending on the force_immediate flag and the partition's type and RMC
    state, this method may attempt increasingly aggressive mechanisms for
    shutting down the OS if initial attempts fail or time out.

    :param part: The LPAR/VIOS wrapper of the instance to power off.
    :param host_uuid: Not used.  Retained for backward compatibility.
    :param force_immediate: One of the Force enum values, defaulting to
                            Force.ON_FAILURE, which behave as follows:
            - Force.TRUE: The force-immediate option is included on the first
                          pass.
            - Force.NO_RETRY: The force-immediate option is not included.  If
                              the power-off fails or times out,
                              VMPowerOffFailure is raised immediately.
            - Force.ON_FAILURE: The force-immediate option is not included on
                                the first pass; but if the power-off fails
                                (including timeout), it is retried with the
                                force-immediate option added.
    :param restart: DEPRECATED: Use a PowerOffOpts instance for add_parms, with
                    restart specified therein.
                    Boolean.  Perform a restart after the power off.  If
                    add_parms is a PowerOffOpts instance, this parameter is
                    ignored.
    :param timeout: Time in seconds to wait for the instance to stop.
    :param add_parms: A power_opts.PowerOffOpts instance; or (deprecated) a
                      dict of parameters to pass directly to the job template.
                      If unspecified, a default PowerOffOpts instance is used,
                      with operation/immediate/restart depending on the
                      force_immediate and restart parameters, and no additional
                      options.
    :raise VMPowerOffFailure: If the operation failed (possibly after retrying)
    :raise VMPowerOffTimeout: If the operation timed out (possibly after
                              retrying).
    """
    opts, legacy = _legacy_power_opts(popts.PowerOffOpts, add_parms)
    if legacy:
        # Decide whether to insist on 'immediate' for OS shutdown.  Do that
        # only if add_parms explicitly included immediate=true.  Otherwise, let
        # soft_detect decide.
        opts.soft_detect(part, immed_if_os=opts.is_immediate or None)
        # Add the restart option if necessary.
        opts.restart(value=restart)

    if force_immediate != Force.ON_FAILURE:
        return _power_off_single(part, opts, force_immediate, timeout)

    # Do the progressive-retry sequence appropriate to the partition type and
    # the force_immediate flag.
    if part.env == bp.LPARType.OS400:
        # The IBMi progression.
        if _pwroff_soft_ibmi_flow(part, opts, timeout):
            return
            # Fall through to VSP hard
    else:
        # The non-IBMi progression.
        if _pwroff_soft_standard_flow(part, opts, timeout):
            return
            # Fall through to VSP hard

    # If we got here, force_immediate == ON_FAILURE, so fall back to VSP hard.
    # Let this one finish or raise.
    # ==> VSP hard
    LOG.warning(_("VSP hard shutdown with default timeout.  Partition: %s"),
                part.name)
    PowerOp.stop(part, opts.vsp_hard())



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

import mock

from pypowervm.tests.test_utils import pvmhttp
import pypowervm.wrappers.http_error as he

HTTPRESP_FILE = "fake_httperror.txt"

MSG = ('Unexpected error occurred while fetching Cluster/SSP '
       'information : 9999-99Z*2125D4A/1 : Unable to send com'
       'mand to VIOS at this moment. VIOS 1*9999-99Z*2125D4A '
       'is busy processing some other request. Please retry t'
       'he operation after sometime.')
MSG2 = ('Error occurred while querying for Adapter from VIOS vios1 with ID 2 '
        'in System 9119-MHE*1085B07 -  The system is currently too busy to '
        'complete the specified request. Please retry the operation at a '
        'later time. If the operation continues to fail, check the error log '
        'to see if the filesystem is full.')
REASON_CODE = 'Unknown internal error.'


class TestHttpError(unittest.TestCase):

    def setUp(self):
        super(TestHttpError, self).setUp()

        self.http_error = pvmhttp.load_pvm_resp(HTTPRESP_FILE)
        self.assertIsNotNone(self.http_error,
                             "Could not load %s " %
                             HTTPRESP_FILE)

    def test_wrap(self):
        wrap = he.HttpError.wrap(self.http_error.response.entry)
        self.assertEqual(wrap.message, MSG)
        self.assertEqual(wrap.status, 500)
        self.assertEqual(wrap.reason_code, REASON_CODE)

        self.assertTrue(wrap.is_vios_busy())
        # Ensure it's checking for 500 only.
        with mock.patch.object(wrap, '_get_val_int', return_value=555):
            self.assertFalse(wrap.is_vios_busy())
        # Ensure it's checking for 'VIOS' string.
        with mock.patch.object(wrap, '_get_val_str', return_value='other'):
            self.assertFalse(wrap.is_vios_busy())
        # Ensure it finds 'VIOS' but not the other string.
        with mock.patch.object(wrap, '_get_val_str', return_value='VIOS xxx'):
            self.assertFalse(wrap.is_vios_busy())

        # Ensure it finds 'HSCL' return code.
        msg_string = 'msg HSCL3205 msg'
        with mock.patch.object(wrap, '_get_val_str', return_value=msg_string):
            self.assertTrue(wrap.is_vios_busy())

        msg_string = 'msg VIOS0014 msg'
        with mock.patch.object(wrap, '_get_val_str', return_value=msg_string):
            self.assertTrue(wrap.is_vios_busy())

        # Ensure we find the new message
        with mock.patch.object(wrap, '_get_val_str', return_value=MSG2):
            self.assertTrue(wrap.is_vios_busy())



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

import pypowervm.wrappers.cluster as clust
from pypowervm.wrappers import job
import pypowervm.wrappers.logical_partition as lpar
import pypowervm.wrappers.managed_system as ms
import pypowervm.wrappers.network as net
import pypowervm.wrappers.shared_proc_pool as spp
import pypowervm.wrappers.storage as stor
import pypowervm.wrappers.vios_file as vf
import pypowervm.wrappers.virtual_io_server as vios


class TestSearch(unittest.TestCase):

    expected_search_keys = {
        clust.Cluster: dict(name='ClusterName'),
        job.Job: None,
        lpar.LPAR: dict(name='PartitionName', id='PartitionID'),
        ms.System: None,
        net.NetBridge: None,
        net.VNet: None,
        net.CNA: None,
        net.VSwitch: None,
        spp.SharedProcPool: None,
        stor.SSP: dict(name='StoragePoolName'),
        stor.VG: None,
        vf.File: None,
        vios.VIOS: dict(name='PartitionName', id='PartitionID'),
    }

    def test_all_search_keys(self):
        for wcls in list(self.expected_search_keys):
            sk = self.expected_search_keys[wcls]
            if sk is None:
                self.assertFalse(hasattr(wcls, 'search_keys'))
            else:
                self.assertEqual(sk, wcls.search_keys)

if __name__ == '__main__':
    unittest.main()



# Copyright 2014, 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import pypowervm.tests.test_utils.test_wrapper_abc as twrap
import pypowervm.wrappers.shared_proc_pool as spp

SHRPROC_HTTPRESP_FILE = "shrprocpool.txt"


class TestShrPrcPoolTestCase(twrap.TestWrapper):

    file = 'shrprocpool.txt'
    wrapper_class_to_test = spp.SharedProcPool

    def test_validate_attribues(self):
        # First element from the feed is the default pool
        self.assertEqual('DefaultPool', self.dwrap.name)
        self.assertEqual(0, self.dwrap.id)
        self.assertEqual(0, self.dwrap.curr_rsrv_proc_units)
        self.assertTrue(self.dwrap.is_default)
        self.assertEqual(0, self.dwrap.max_proc_units)
        self.assertEqual(0, self.dwrap.pend_rsrv_proc_units)
        self.assertEqual(0, self.dwrap.avail_proc_units)

        # The second pool is non-default.
        n_spp = self.entries[1]
        self.assertEqual('SharedPool01', n_spp.name)
        self.assertEqual(1, n_spp.id)
        self.assertEqual(5.35, n_spp.curr_rsrv_proc_units)
        self.assertFalse(n_spp.is_default)
        self.assertEqual(10.25, n_spp.max_proc_units)
        self.assertEqual(6.15, n_spp.pend_rsrv_proc_units)
        self.assertEqual(5.05, n_spp.avail_proc_units)

    def test_setters(self):
        self.dwrap.name = 'new'
        self.assertEqual('new', self.dwrap.name)

        self.dwrap.max_proc_units = 5.5
        self.assertEqual(5.5, self.dwrap.max_proc_units)

        self.dwrap.pend_rsrv_proc_units = 4.3
        self.assertEqual(4.3, self.dwrap.pend_rsrv_proc_units)



# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from pypowervm import util


def udid_to_scsi_mapping(vios_w, udid, lpar_id, ignore_orphan=True):
    """Finds the SCSI mapping (if any) for a given backing storage udid.

    This is a helper method that will parse through a given VIOS wrapper
    (retrieved with pypowervm.const.XAG.VIO_SMAP) and will find the client
    SCSI mapping for a given backing storage element (LU, PV, LV, VOpt).

    :param vios_w: The Virtual I/O Server wrapper.  Should have the Storage
                   and SCSI mapping XAG associated with it.
    :param udid: The volume's udid.
    :param lpar_id: The LPARs 'short' id.
    :param ignore_orphan: (Optional, Default: True) If set to True, any orphan
                          SCSI mappings (those with no client adapter) will be
                          ignored.
    :return: The first matching SCSI mapping (or None).
    """
    for scsi_map in vios_w.scsi_mappings:
        # No backing storage, then ignore.
        if not scsi_map.backing_storage:
            continue

        # If there is not a client adapter, it isn't attached fully.
        if not scsi_map.client_adapter and ignore_orphan:
            continue

        # Is it for the right LPAR?  (The server adapter is present even if
        # it's an orphan.)
        if lpar_id != scsi_map.server_adapter.lpar_id:
            continue

        if scsi_map.backing_storage.udid == udid:
            return scsi_map

    return None


def c_wwpn_to_vfc_mapping(vios_w, c_wwpn):
    """Finds the vFC mapping (if any) for a given client WWPN.

    This is a helper method that will parse through a given VIOS wrapper
    (retrieved with pypowervm.const.XAG.VIO_FMAP) and will find the client vFC
    mapping for that WWPN.

    :param vios_w: The Virtual I/O Server wrapper.  Should have
                   pypowervm.const.XAG.VIO_FMAP associated with it.
    :param c_wwpn: One of the client's WWPNs.
    :return: The vFC mapping (or None)
    """
    wwpn = util.sanitize_wwpn_for_api(c_wwpn)
    for vfc_map in vios_w.vfc_mappings:
        # If there is not a client adapter, it isn't properly attached.
        if not vfc_map.client_adapter:
            continue

        if wwpn in vfc_map.client_adapter.wwpns:
            return vfc_map

    return None
